from collections.abc import Iterator
from threading import Thread

import pytest

from karabo.bound import EventLoop


@pytest.fixture(scope="module")
def eventLoop() -> Iterator[EventLoop]:
    thread: Thread = Thread(target=EventLoop.work, daemon=True)
    thread.start()
    try:
        yield EventLoop
    finally:
        EventLoop.stop()
        thread.join(timeout=10)
        if thread.is_alive():
            raise RuntimeError("Thread still alive after joining ...")



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

import threading
from time import sleep

import pytest

from karabo.bound import (
    Broker, Connection, EventLoop, Hash, Schema, SignalSlotable, fullyEqual)

debugFlag = False
lock = threading.Lock()

# For debugging to get exceptions printed that are caugth by C++ event loop:
# from karabo.bound import Logger
# Logger.configure()
# Logger.useConsole()


@pytest.fixture(scope="module")
def eventLoop():
    loop_thread = threading.Thread(target=EventLoop.work)
    loop_thread.start()
    # test_asynch_write_read needs an extra thread if readAsyncXxx would be
    # called before writeAsyncXxx
    EventLoop.addThread()

    yield  # now all (since scope="module") tests are executed

    EventLoop.stop()
    loop_thread.join(timeout=10)
    assert not loop_thread.is_alive()


def test_broker():

    s = Schema()
    Broker.expectedParameters(s)
    assert s.has("brokers")

    brokerType = Broker.brokerTypeFromEnv()
    assert brokerType in Broker.getRegisteredClasses()

    # Create a Broker via SignalSlotable to test the only (besides from some
    # macro) non-static member function that is exposed to Python:
    sigSlot = SignalSlotable("brokerCreator", Hash(), 60, Hash())
    broker = sigSlot.getConnection()
    assert type(broker) is Broker
    brokerProtocol = brokerType
    assert broker.getBrokerUrl().startswith(brokerProtocol + "://")


# Could become a fixture that requires eventLoop fixture, but then I do not
# know how to pass the classes
def setup_server_client():
    """
    Create server and client connected to each other, using Connection and Hash
    classes as passed by argument.
    """
    # Alice as server, bob as client. Server cannot be started synchronously
    # since then the port that the client should connect to is not available.

    aliceConn = Connection.create("Tcp", Hash("type", "server"))
    alice = None  # to become the Channel

    connectedCond = threading.Condition()

    def onConnect(ec, channel):
        nonlocal alice
        alice = channel
        with connectedCond:
            connectedCond.notify()

    aPort = aliceConn.startAsync(onConnect)

    bobCfg = Hash("type", "client", "hostname", "localhost", "port", aPort)
    bobConn = Connection.create("Tcp", bobCfg)
    with connectedCond:
        bob = bobConn.start()  # bob is the Channel
        connectedCond.wait(timeout=10)  # wait until notified

    assert alice is not None
    aliceConn2 = alice.getConnection()
    assert type(aliceConn) is Connection
    assert aliceConn2 is aliceConn

    # Better return (and thus keep alive) the connections as well:
    # C++ TcpChannel keeps only weak_ptr to its TcpConnection.
    return alice, aliceConn, bob, bobConn


def test_synch_write_read(eventLoop):

    alice, aliceConn, bob, bobConn = setup_server_client()

    alice.write("messageStr")  # str
    readMsg = bob.readStr()
    assert readMsg == "messageStr"

    alice.write(b"messageBytes")  # bytes
    readMsg = bob.readStr()
    assert readMsg == "messageBytes"

    alice.write(bytearray("messageBytearray", "utf-8"))  # bytearray
    readMsg = bob.readStr()
    assert readMsg == "messageBytearray"

    alice.write(Hash("hash", "message"))  # Hash
    readMsg = bob.readHash()
    assert fullyEqual(readMsg, Hash("hash", "message"))

    with pytest.raises(RuntimeError) as excinfo:
        alice.write(1)  # not supported to write int
        assert "Python Exception: Not supported type" in str(excinfo.value)

    alice.write(Hash("str", "header"), "messageStr")  # Hash, str
    readHeader, readMsg = bob.readHashStr()
    assert fullyEqual(readHeader, Hash("str", "header"))
    assert readMsg == "messageStr"

    alice.write(Hash("bytes", "header"), b"messageBytes")  # Hash, bytes
    readHeader, readMsg = bob.readHashStr()
    assert fullyEqual(readHeader, Hash("bytes", "header"))
    assert readMsg == "messageBytes"

    alice.write(Hash("bytearray", "header"),
                bytearray("messageBytearray", "utf-8"))  # bytearray
    readHeader, readMsg = bob.readHashStr()
    assert fullyEqual(readHeader, Hash("bytearray", "header"))
    assert readMsg == "messageBytearray"

    alice.write(Hash("a", "header"), Hash("hash", "message"))  # Hash, Hash
    readHeader, readMsg = bob.readHashHash()
    assert fullyEqual(readHeader, Hash("a", "header"))
    assert fullyEqual(readMsg, Hash("hash", "message"))

    with pytest.raises(RuntimeError) as excinfo:
        alice.write(Hash("a", "header"), 1)  # not supported
        assert "Python Exception: Not supported type" in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        alice.write(1, "message")  # not supported header type
        assert ("TypeError:  write(): incompatible function arguments."
                in str(excinfo.value))


def test_asynch_write_read(eventLoop):
    # Test asynchronous write and read methods.

    alice, aliceConn, bob, bobConn = setup_server_client()

    ###################################################################
    # Write and read str
    ###################################################################

    ecWrite = None
    channelWrite = None
    conditionWrite = threading.Condition()

    def writeComplete(ec, channel):
        nonlocal ecWrite, channelWrite
        ecWrite = ec
        channelWrite = channel
        with conditionWrite:
            conditionWrite.notify()

    alice.writeAsyncStr("messageStr", writeComplete)
    with conditionWrite:
        conditionWrite.wait(timeout=10)
    assert ecWrite.value() == 0
    assert alice is channelWrite

    ecRead = None
    channelRead = None
    msgRead = None
    conditionRead = threading.Condition()

    def onRead1(ec, channel, msg):
        nonlocal ecRead, channelRead, msgRead
        ecRead = ec
        channelRead = channel
        msgRead = msg
        with conditionRead:
            conditionRead.notify()

    bob.readAsyncStr(onRead1)
    with conditionRead:
        conditionRead.wait(timeout=10)
    assert ecRead.value() == 0
    assert bob is channelRead
    assert msgRead == "messageStr"

    ###################################################################
    # Write bytes (and read as str)
    ###################################################################

    # Reset to re-use writeComplete and onRead1
    ecWrite = channelWrite = None
    ecRead = channelRead = msgRead = None

    alice.writeAsyncStr(b"messageStr", writeComplete)
    with conditionWrite:
        conditionWrite.wait(timeout=10)
    assert ecWrite.value() == 0
    assert alice is channelWrite

    bob.readAsyncStr(onRead1)
    with conditionRead:
        conditionRead.wait(timeout=10)
    assert ecRead.value() == 0
    assert bob is channelRead
    assert msgRead == "messageStr"

    ###################################################################
    # Write bytearray (and read as str)
    ###################################################################
    # Reset to re-use writeComplete and onRead1
    ecWrite = channelWrite = ecRead = None
    channelRead = msgRead = None

    alice.writeAsyncStr(bytearray("messageStr", encoding="utf8"),
                        writeComplete)
    with conditionWrite:
        conditionWrite.wait(timeout=10)
    assert ecWrite.value() == 0
    assert alice is channelWrite

    bob.readAsyncStr(onRead1)

    with conditionRead:
        conditionRead.wait(timeout=10)
    assert ecRead.value() == 0
    assert bob is channelRead
    assert msgRead == "messageStr"

    ###################################################################
    # Write and read Hash
    ###################################################################
    # Reset to re-use writeComplete and onRead1
    ecWrite = channelWrite = None
    ecRead = channelRead = msgRead = None

    alice.writeAsyncHash(Hash("hash", "message"), writeComplete)
    with conditionWrite:
        conditionWrite.wait(timeout=10)
    assert ecWrite.value() == 0
    assert alice is channelWrite

    bob.readAsyncHash(onRead1)
    with conditionRead:
        conditionRead.wait(timeout=10)
    assert ecRead.value() == 0
    assert bob is channelRead
    assert fullyEqual(msgRead, Hash("hash", "message"))

    ###################################################################
    # Write and read Hash, string
    ###################################################################

    # Reset to re-use writeComplete
    ecWrite = channelWrite = None

    # Need an onRead2 for header/body messages
    ecRead = channelRead = headerRead = msgRead = None

    def onRead2(ec, channel, header, msg):
        nonlocal ecRead, channelRead, headerRead, msgRead
        ecRead = ec
        channelRead = channel
        headerRead = header
        msgRead = msg
        with conditionRead:
            conditionRead.notify()

    alice.writeAsyncHashStr(Hash("header", "message"), "body",
                            writeComplete)
    with conditionWrite:
        conditionWrite.wait(timeout=10)
    assert ecWrite.value() == 0
    assert alice is channelWrite

    bob.readAsyncHashStr(onRead2)

    with conditionRead:
        conditionRead.wait(timeout=10)
    assert ecRead.value() == 0
    assert bob is channelRead
    assert fullyEqual(headerRead, Hash("header", "message"))
    assert msgRead == "body"

    ###################################################################
    # Write and read Hash, bytes
    ###################################################################
    # Reset to re-use writeComplete and onRead2
    ecWrite = channelWrite = None
    ecRead = channelRead = headerRead = msgRead = None

    alice.writeAsyncHashStr(Hash("header", "message b"), b"bytes",
                            writeComplete)
    with conditionWrite:
        conditionWrite.wait(timeout=10)
    assert ecWrite.value() == 0
    assert alice is channelWrite

    bob.readAsyncHashStr(onRead2)

    with conditionRead:
        conditionRead.wait(timeout=10)
    assert ecRead.value() == 0
    assert bob is channelRead
    assert fullyEqual(headerRead, Hash("header", "message b"))
    assert msgRead == "bytes"

    ###################################################################
    # Write and read Hash, bytearray
    ###################################################################
    # Reset to re-use writeComplete and onRead2
    ecWrite = channelWrite = None
    ecRead = channelRead = headerRead = msgRead = None

    alice.writeAsyncHashStr(Hash("header", "message b2"),
                            bytearray("bytearray", encoding="utf8"),
                            writeComplete)
    with conditionWrite:
        conditionWrite.wait(timeout=10)
    assert ecWrite.value() == 0
    assert alice is channelWrite

    bob.readAsyncHashStr(onRead2)

    with conditionRead:
        conditionRead.wait(timeout=10)
    assert ecRead.value() == 0
    assert bob is channelRead
    assert fullyEqual(headerRead, Hash("header", "message b2"))
    assert msgRead == "bytearray"

    ###################################################################
    # Write and read Hash, Hash
    ###################################################################
    # Reset to re-use writeComplete and onRead2
    ecWrite = channelWrite = None
    ecRead = channelRead = headerRead = msgRead = None

    alice.writeAsyncHashHash(Hash("header", "message a"),
                             Hash("body", "message b"),
                             writeComplete)
    with conditionWrite:
        conditionWrite.wait(timeout=10)
    assert ecWrite.value() == 0
    assert alice is channelWrite

    bob.readAsyncHashHash(onRead2)

    with conditionRead:
        conditionRead.wait(timeout=10)
    assert ecRead.value() == 0
    assert bob is channelRead
    assert fullyEqual(headerRead, Hash("header", "message a"))
    assert fullyEqual(msgRead, Hash("body", "message b"))


def DEBUG(s):
    if debugFlag:
        lock.acquire()
        print(s)
        lock.release()


class Server:
    def __init__(self, Connection, Hash):
        # create connection object
        self.conn = Connection.create("Tcp", Hash("type", "server", "port", 0))
        # register connect handler for incoming connections
        self.port = self.conn.startAsync(self.onConnect)
        DEBUG(f"\nSRV listening port : {self.port}")

    def onError(self, ec, channel):
        DEBUG(f"SRV onError #{ec.value()} => {ec.message()}")
        channel.close()
        DEBUG("SRV channel closed")

    def onWriteHashHashComplete(self, ec, channel):
        DEBUG("SRV onWriteHashComplete entry.")
        if ec.value() != 0:
            self.onError(ec, channel)
            return
        # register for reading new Hash/Hash message
        channel.readAsyncHashHash(self.onReadHashHash)
        DEBUG("SRV onWriteHashHashComplete exit.")

    def onReadHashHash(self, ec, channel, h, b):
        DEBUG(f"SRV onReadHashHash entry: #{ec.value()} => {ec.message()}")
        if ec.value() != 0:
            self.onError(ec, channel)
            return
        # process the message
        b["server"] = "SERVER: message processed!"
        # send back to client
        channel.writeAsyncHashHash(h, b, self.onWriteHashHashComplete)
        DEBUG("SRV onReadHashHash exit.")

    def onWriteHashStrComplete(self, ec, channel):
        DEBUG("SRV onWriteHashStrComplete entry.")
        if ec.value() != 0:
            self.onError(ec, channel)
            return
        # register for reading new Hash/Hash message
        channel.readAsyncHashHash(self.onReadHashHash)
        DEBUG("SRV onWriteHashStrComplete exit.")

    def onReadHashStr(self, ec, channel, h, s):
        DEBUG(f"SRV onReadHashStr entry: #{ec.value()} => {ec.message()}")
        if ec.value() != 0:
            self.onError(ec, channel)
            return
        # send back to client
        channel.writeAsyncHashStr(h, s, self.onWriteHashStrComplete)
        DEBUG("SRV onReadHashStr exit.")

    def onWriteHashComplete(self, ec, channel):
        DEBUG("SRV onWriteHashComplete entry.")
        if ec.value() != 0:
            self.onError(ec, channel)
            return
        # register for reading new Hash/Str message
        channel.readAsyncHashStr(self.onReadHashStr)
        DEBUG("SRV onWriteHashComplete exit.")

    def onReadHash(self, ec, channel, h):
        DEBUG(f"SRV onReadHash entry: #{ec.value()} => {ec.message()}")
        if ec.value() != 0:
            self.onError(ec, channel)
            return
        # send back to client
        channel.writeAsyncHash(h, self.onWriteHashComplete)
        DEBUG("SRV onReadHash exit.")

    def onWriteStrComplete(self, ec, channel):
        DEBUG("SRV onWriteStrComplete entry.")
        if ec.value() != 0:
            self.onError(ec, channel)
            return
        # register for reading new Hash message
        channel.readAsyncHash(self.onReadHash)
        DEBUG("SRV onWriteStrComplete exit.")

    def onReadStr(self, ec, channel, s):
        DEBUG(f"SRV onReadStr entry: #{ec.value()} => {ec.message()}")
        if ec.value() != 0:
            self.onError(ec, channel)
            return
        channel.writeAsyncStr(s, self.onWriteStrComplete)
        DEBUG("SRV onReadStr exit.")

    def onConnect(self, ec, channel):
        DEBUG(f"SRV onConnect entry: #{ec.value()} -> {ec.message()}")
        if ec.value() != 0:
            self.onError(ec, channel)
            return
        # register read handler for incoming Str messages
        channel.readAsyncStr(self.onReadStr)
        DEBUG("SRV onConnect exit.")

    # this method stops server
    def stop(self):
        self.conn.stop()


def test_tcp_client_server(eventLoop):
    counter = 0
    count_max = 10
    server = Server(Connection, Hash)
    DEBUG("CLN: Server started...")

    failed = None
    done = False

    def onError(ec, channel):
        nonlocal done
        DEBUG("CLN: onError")
        channel.close()
        server.stop()
        done = True

    def onReadHashHash(ec, channel, header, body):
        nonlocal failed, counter, done
        counter += 1
        DEBUG(f"CLN onReadHashHash entry: #{ec.value()} => {ec.message()}")
        if ec.value() != 0:
            onError(ec, channel)
            return

        # Inspect received Hash ...
        try:
            assert len(body) == 4
            assert body['server'] == "SERVER: message processed!"
            assert body['a.b.c'] == 1
            assert body['x.y.z'] == [1, 2, 3, 4, 5]
            assert body['d.abc'] == 'rubbish'
        except BaseException as exc:
            failed = f'Hash inspection failed: {exc}'

        if counter >= count_max:
            channel.close()
            server.stop()
            DEBUG("CLN: DONE")
            done = True
        else:
            # Next message ...
            body = Hash(
                "a.b.c", 1,
                "x.y.z", [1, 2, 3, 4, 5],
                "d", Hash("abc", "rubbish"))
            header = Hash("par1", 12)
            channel.writeAsyncHashHash(header, body, onWriteHashHashComplete)
        DEBUG(f"CLN onReadHashHash exit.\t\tCOUNT = {counter}\n")

    def onWriteHashHashComplete(ec, channel):
        DEBUG("CLN onWriteHashHashComplete entry.")
        if ec.value() != 0:
            onError(ec, channel)
            return

        with pytest.raises(RuntimeError):
            channel.readAsyncHashHash(None)

        channel.readAsyncHashHash(onReadHashHash)
        DEBUG("CLN onWriteHashHashComplete exit")

    def onReadHashStr(ec, channel, h, s):
        nonlocal failed
        nonlocal counter
        counter += 1
        DEBUG(f"CLN onReadHashStr entry: #{ec.value()} => {ec.message()}")
        if ec.value() != 0:
            onError(ec, channel)
            return

        try:
            assert s == "some message"
        except BaseException as exc:
            failed = f'Message inspection failed: {exc}'
            return

        # Next message ...
        body = Hash(
            "a.b.c", 1,
            "x.y.z", [1, 2, 3, 4, 5],
            "d", Hash("abc", "rubbish"))
        header = Hash("par1", 12)

        with pytest.raises(RuntimeError):
            channel.writeAsyncHashHash(header, body, None)

        channel.writeAsyncHashHash(header, body, onWriteHashHashComplete)
        DEBUG(f"CLN onReadHashStr exit.\t\t\tCOUNT = {counter}\n")

    def onWriteHashStrComplete(ec, channel):
        DEBUG("CLN onWriteHashStrComplete entry.")
        if ec.value() != 0:
            onError(ec, channel)
            return

        with pytest.raises(RuntimeError):
            channel.readAsyncHashStr(None)

        channel.readAsyncHashStr(onReadHashStr)
        DEBUG("CLN onWriteHashStrComplete exit")

    def onReadHash(ec, channel, body):
        nonlocal failed
        nonlocal counter
        counter += 1
        DEBUG(f"CLN onReadHash entry: #{ec.value()} => {ec.message()}")
        if ec.value() != 0:
            onError(ec, channel)
            return

        try:
            assert len(body) == 1
            assert body['a'] == 12
        except BaseException as exc:
            failed = f'Message inspection failed: {exc}'
            return

        h = Hash('a', 1)
        s = "some message"
        channel.writeAsyncHashStr(h, s, onWriteHashStrComplete)
        DEBUG(f"CLN onReadHash exit.\t\t\tCOUNT = {counter}\n")

    def onWriteHashComplete(ec, channel):
        DEBUG("CLN onWriteHashComplete entry.")
        if ec.value() != 0:
            onError(ec, channel)
            return

        with pytest.raises(RuntimeError):
            channel.readAsyncHash(None)

        channel.readAsyncHash(onReadHash)
        DEBUG("CLN onWriteHashComplete exit")

    def onReadStr(ec, channel, s):
        nonlocal failed
        nonlocal counter
        counter += 1
        DEBUG(f"CLN onReadStr entry: #{ec.value()} => {ec.message()}")
        if ec.value() != 0:
            onError(ec, channel)
            return

        try:
            assert s == 'Hello'
        except BaseException as exc:
            failed = f'Message inspection failed: {exc}'
            return

        h = Hash("a", 12)
        channel.writeAsyncHash(h, onWriteHashComplete)
        DEBUG(f"CLN onReadStr exit.\t\t\tCOUNT = {counter}\n")

    def onWriteStrComplete(ec, channel):
        DEBUG("CLN onWriteStrComplete entry.")
        if ec.value() != 0:
            onError(ec, channel)
            return

        with pytest.raises(RuntimeError):
            channel.readAsyncStr(None)

        channel.readAsyncStr(onReadStr)
        DEBUG("CLN onWriteStrComplete exit")

    def onConnect(ec, channel):
        nonlocal counter
        counter = 0
        DEBUG(f"CLN onConnect entry: #{ec.value()} => {ec.message()}")
        if ec.value() != 0:
            onError(ec, channel)
            return

        with pytest.raises(RuntimeError):
            channel.writeAsyncStr("Tschüß", None)

        channel.writeAsyncStr("Hello", onWriteStrComplete)
        DEBUG(f"CLN onConnect exit.\t\t\tCOUNT = {counter}\n")

    # Asynchronous TCP client
    # create client connection object
    DEBUG("CLN create Connection and start it")
    client = Connection.create(
        "Tcp", Hash("type", "client", "hostname", "localhost",
                    "port", server.port))
    with pytest.raises(RuntimeError):
        client.startAsync(None)
    client.startAsync(onConnect)
    DEBUG("CLN started Connection")

    # Wait until 'done' (could be done nice with condition variable or so...)
    timeout = 10
    while timeout > 0:
        if done or failed is not None:
            break
        timeout -= 0.01
        sleep(0.01)

    assert failed is None
    assert counter == count_max

    DEBUG("CLN EXIT")



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

import threading
import traceback
from collections.abc import Callable
from typing import Self


class Worker(threading.Thread):
    def __init__(self,
                 callback: Callable[[], None] | None = None,
                 timeout: int | float = 1000,
                 single_shot: bool = False,
                 daemon: bool = True) -> None:
        """Constructs the Worker thread, that is by default a daemon thread.

        Please note that daemon threads may not release resources
        properly when stopped abruptly.

        Note: `timeout` is in milliseconds.
        """
        super().__init__(daemon=daemon)
        self.callback: Callable[[], None] | None = callback
        self.single_shot: bool = single_shot
        self.running: bool = False
        self.aborted: bool = False
        self.suspended: bool = False
        self.cv: threading.Condition = threading.Condition()
        if not timeout > 0:
            raise RuntimeError("Timeout needs to be larger than zero.")
        self.timeout: int | float = timeout

    def set(self, callback: Callable[[], None],
            timeout: int | float = 1000) -> None:
        if self.is_running():
            raise RuntimeError("Cannot change parameters while running.")
        self.callback = callback
        if not timeout > 0:
            raise RuntimeError("Timeout needs to be larger than zero.")
        self.timeout = timeout

    def setSingleShot(self, single_shot: bool) -> None:
        self.single_shot = single_shot

    def setTimeout(self, timeout: int | float = 1000) -> None:
        if not timeout > 0:
            raise RuntimeError("Timeout needs to be larger than zero.")
        self.timeout = timeout

    def is_running(self) -> bool:
        return self.running

    def run(self) -> None:
        self.running = True
        self.aborted = False
        self.suspended = False
        try:
            assert self.callback is not None, "Callback cannot be None"
            while not self.aborted:
                if not self.running:
                    break

                if self.suspended:
                    with self.cv:
                        while self.suspended:
                            self.cv.wait()
                        if self.aborted or not self.running:
                            break
                    continue

                with self.cv:
                    self.cv.wait(self.timeout / 1000.0)

                if self.suspended:
                    continue

                if self.running and self.callback:
                    self.callback()
                    if self.single_shot:
                        break
        except Exception:
            traceback.print_exc()
        finally:
            self.running = False

    def start(self) -> Self:
        if not self.running:
            self.suspended = False
            super().start()

        if self.suspended:
            with self.cv:
                self.suspended = False
                self.cv.notify()

        return self

    def stop(self) -> Self:
        if self.running:
            with self.cv:
                self.running = False
                self.suspended = False
                self.cv.notify()
        return self

    def abort(self) -> Self:
        self.aborted = True
        self.running = False
        if self.suspended:
            with self.cv:
                self.suspended = False
                self.cv.notify()
        return self

    def pause(self) -> None:
        if not self.suspended:
            with self.cv:
                self.suspended = True
                self.cv.notify()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import copy

import numpy as np
import pytest

from karabo.bound import (
    Hash, HashMergePolicy, HashNode as Node, Types, VectorHash, fullyEqual,
    similar)

# Test "clear", "empty", "getKeys", "keys", "getValues", "values",
#      "getPaths", "paths", "set", "__setitem__", "setAs", "get",
#      "__getitem__", "has", "__contains__", "erase", "__iter__",
#      "__delitem__", "erasePath", "__len__", "__bool__", "getAs",
#      "getType", "merge", "__iadd__", "__isub__", "similar",
#      "fullyEqual", "isType", "flatten", "unflatten", "find",
#      "setNode", "getNode", "hasAttribute", "getAttribute",
#      "getAttributeAs", "getAttributes", "copyAttributes",
#      "setAttribute", "setAttributes", "__copy__",
#      "__deepcopy__"


def test_constructor():
    h = Hash()

    # Test "__len__"
    assert len(h) == 0

    # Test "empty"
    assert h.empty()

    # Test "__bool__"
    assert not h
    # It is allowed any number of key/value pairs to be defined
    h = Hash('a', 1, 'b', 2, 'c', 3, 'd', 4, 'e', 5, 'f', 6)
    assert h
    # For further populating, use set ...
    h.set('g', 7)
    assert len(h) == 7
    assert not h.empty()

    # Test "clear"
    h.clear()
    assert h.empty()

    # Test constructors
    # There are no restrictions in number of key/value pairs
    h = Hash('a', 1, 'b', 2, 'c', 3, 'd', 4, 'e', 5, 'f', 6, 'g', 7, 'h', 8)
    assert h
    # Define python dictionary...
    d = {'a': 1, 'b': {'c': 12}, 'c': {'e': {'f': "abrakadabra"}}}
    h = Hash(d)
    assert h
    assert "c.e.f" in h
    assert h.get('b.c') == 12
    # Cannot construct Node object in Python ...
    with pytest.raises(TypeError):
        print(Node())


def test_getset():
    # Test "get" .... default value
    h = Hash()
    assert h.get("missing") is None
    assert h.get("missing", default="value") == "value"

    # Test "set"
    h.set('a', "some text")
    h.set('b.c.d.e.f', 2.718281828)

    # Test "__setitem__"
    h['w.r.a.p'] = 123456789999
    h['c.c.c.c'] = True
    h['p.e.r.f'] = [1.1, 2.2, 3.3]

    # Test "get"
    assert h.get('a') == "some text"
    assert h.get('p.e.r.f') == [1.1, 2.2, 3.3]

    # Test "__getItem__"
    assert h['b.c.d.e.f'] == 2.718281828
    assert h['c.c.c.c'] is True

    # Test "getType"
    assert h.getType('b.c.d.e.f') == Types.DOUBLE
    assert h.getType('p.e.r.f') == Types.VECTOR_DOUBLE

    # Test "isType"
    assert h.isType('a', Types.STRING)
    assert h.isType('c.c.c.c', Types.BOOL)
    assert h.isType('a', "STRING")
    assert h.isType('c.c.c.c', "BOOL")

    # Test "has"
    assert h.has('p.e.r.f')
    assert not h.has("miss.ing")

    # Test "keys"
    assert h.keys() == ['a', 'b', 'w', 'c', 'p']

    # Test "paths"
    assert h.paths() == ['a', 'b.c.d.e.f', 'w.r.a.p', 'c.c.c.c', 'p.e.r.f']
    # Test "getKeys"
    assert h.getKeys() == ['a', 'b', 'w', 'c', 'p']

    # Test "getPaths"
    assert h.getPaths() == ['a', 'b.c.d.e.f', 'w.r.a.p', 'c.c.c.c', 'p.e.r.f']
    # Test C++-ish "getKeys"
    res = []
    h.getKeys(res)
    assert res == ['a', 'b', 'w', 'c', 'p']

    # Test C++-ish "getPaths"
    res = []
    h.getPaths(res)
    assert res == ['a', 'b.c.d.e.f', 'w.r.a.p', 'c.c.c.c', 'p.e.r.f']
    # Test "values"
    assert h.values() == ["some text",
                          Hash('c.d.e.f', 2.718281828),
                          Hash('r.a.p', 123456789999),
                          Hash('c.c.c', True),
                          Hash('e.r.f', [1.1, 2.2, 3.3])]
    # Test "getValues"
    assert h.getValues() == ["some text",
                             Hash('c.d.e.f', 2.718281828),
                             Hash('r.a.p', 123456789999),
                             Hash('c.c.c', True),
                             Hash('e.r.f', [1.1, 2.2, 3.3])]

    # Test "flatten"
    g = Hash()
    h.flatten(g)
    assert g.keys() == ['a', 'b.c.d.e.f', 'w.r.a.p',
                        'c.c.c.c', 'p.e.r.f']
    assert g.keys() == g.paths()
    # ERROR: assert 'b.c.d.e.f' in g)
    assert g.has('b.c.d.e.f', '/')
    assert not g.has('b.c.d', '/')
    assert g.values() == ["some text", 2.718281828,
                          123456789999, True, [1.1, 2.2, 3.3]
                          ]
    # NOTE: flatten Hash has keys containing '.' considered as default
    # separator.  The 'get with default separator will fail.  To get
    # the value, one can use another separator like
    assert g.get('b.c.d.e.f', '/') == 2.718281828
    assert g.get('w.r.a.p', '|') == 123456789999

    # Test "unflatten"
    f = Hash()
    g.unflatten(f)
    assert f == h  # '==' is 'similar'
    assert similar(f, h)
    assert fullyEqual(f, h)
    g.clear()
    assert not g

    # Test "erase"
    f.erase('b.c.d.e')
    assert 'b' in f
    assert 'b.c.d.e' not in f
    assert 'b.c.d' in f
    assert f.getType('b.c.d') == Types.HASH
    assert f['b.c.d'] == Hash()

    # Test "erasePath"
    assert f['p.e.r.f'] == [1.1, 2.2, 3.3]
    f.erasePath('p.e.r')
    assert 'p.e.r' not in f
    assert 'p.e' not in f
    assert 'p' not in f

    # Test "__delitem__"
    del f['w.r']
    assert 'w.r' not in f
    assert 'w' in f
    assert f['w'] == Hash()
    assert h['w'] == Hash('r.a.p', 123456789999)

    del f

    # Test "__iter__"
    assert len(h) == 5
    it = iter(h)

    assert h['a'] == "some text"
    n1 = next(it)
    assert isinstance(n1, Node)
    assert n1.getKey() == 'a'
    assert n1.getValue() == 'some text'
    assert n1.getType() == Types.STRING

    assert h['b.c.d.e.f'] == 2.718281828
    n2 = next(it)
    assert isinstance(n2, Node)
    assert n2.getKey() == 'b'
    assert n2.getValue(), Hash('c.d.e.f', 2.718281828)
    assert n2.getType() == Types.HASH

    assert h['w.r.a.p'] == 123456789999
    n3 = next(it)
    assert isinstance(n3, Node)
    assert n3.getKey() == 'w'
    assert n3.getValue() == Hash('r.a.p', 123456789999)
    assert n3.getType() == Types.HASH

    assert h['c.c.c.c'] is True
    n4 = next(it)
    assert isinstance(n4, Node)
    assert n4.getKey() == 'c'
    assert n4.getValue() == Hash('c.c.c', True)
    assert n4.getType() == Types.HASH

    assert h['p.e.r.f'] == [1.1, 2.2, 3.3]
    n5 = next(it)
    assert isinstance(n5, Node)
    assert n5.getKey() == 'p'
    assert n5.getValue() == Hash('e.r.f', [1.1, 2.2, 3.3])
    assert n5.getType() == Types.HASH

    with pytest.raises(StopIteration):
        next(it)

    # Test "find"
    n = h.find('b.c.d')
    assert n
    assert n.getKey() == 'd'
    n = h.find('b.c.x')
    assert not n

    # Test "get" and/or "__getitem__" while pointing to the middle
    # of the parent tree 'h'
    # The resulting subtree behaves like a reference
    # Source "parent" tree is ...
    assert h['b.c.d.e.f'] == 2.718281828
    assert h.getType('b.c.d.e.f') == Types.DOUBLE
    assert h.getType('b.c.d') == Types.HASH
    # Test "__getitem__"
    g = h['b.c.d']  # reference to the subtree
    assert g['e.f'] == 2.718281828
    assert g.getType('e.f') == Types.DOUBLE
    # set new value via subtree ref.
    g['e.f'] = 'new long string'
    # the type is changed ...
    assert g.getType('e.f') == Types.STRING
    # Check that this is real reference, not a copy...
    assert h['b.c.d.e.f'] == 'new long string'
    assert h.getType('b.c.d.e.f') == Types.STRING

    # Reference from reference?
    # parent tree ... the owner
    h = Hash('a.b.c.d.e.f', 1)
    assert h['a.b.c.d.e.f'] == 1
    # reference to the node 'a.b.c' ... the proxy
    g = h['a.b.c']
    g['d.e.f'] = 55
    assert h['a.b.c.d.e.f'] == 55
    # reference to the node 'd.e' ... the proxy
    t = g['d.e']
    t['f'] = 88
    assert g['d.e.f'] == 88
    assert h['a.b.c.d.e.f'] == 88
    # drop the parent Hash in Python
    del h
    # Attempt to access 'g' or 't' gives possible despite 'h' is gone
    # (lifetime extended via 'keep_alive<...>()' magic)
    assert g['d.e.f'] == 88
    assert t['f'] == 88

    h = Hash("a[0]", Hash("a", 1), "a[1]", Hash("a", 1))
    assert h["a[0].a"] == 1, "Value should be 1"
    assert h["a[1].a"] == 1, "Value should be 1"
    assert h["a"][0]["a"] == 1, "Value should be 1"
    assert h["a"][1]["a"] == 1, "Value should be 1"
    # destroy parent Hash in C++
    h.clear()
    del g  # the subtree is invalidated. SEGFAULT if accessed.
    assert h.empty()

    # Comparison of 2 Hashes using operators: "==" and "!="
    h = Hash('a.b.c', True, 'x.y.z', [1, 2, 3])
    g = copy.copy(h)
    assert h == g
    # Change some values in the copy
    g['a.b.c'] = False
    g['x.y.z'] = [9, 8, 7, 6, 5]
    # BTW, check that above `copy` is a real copy and not a reference
    assert h['a.b.c'] is True
    assert h['x.y.z'] == [1, 2, 3]

    # Changing the value of leaf does not change the equality because
    # in C++ the operators '==' and '!=' use call (erroneously) to
    # 'similar(...)' method which checks only the "structure" equality
    # and does not compare the leaves ...
    assert h == g
    # But if the structure is changed ...
    h['a.b.c'] = 12  # the type is changed from BOOL to INT32
    assert h != g
    # ... or some new key/value was added ...
    g = copy.copy(h)  # restore equality
    g['h.e.l.p'] = 2.78
    assert h != g

    # Test "getNode"
    h = Hash('a', 12)
    n = h.getNode('a')
    assert n.getKey() == 'a'
    assert n.getValue() == 12
    assert n.getType() == Types.INT32

    # Test "setNode"
    f = Hash('b', "text")
    f.setNode(n)
    assert 'b' in f
    assert 'a' in f
    assert f['a'] == 12
    assert n.getValue() == 12
    assert f.getType('a') == Types.INT32

    # test 'items()'...
    h = Hash('a.b', 12, 'c.d', 23, 'e', 65, 'f.g.m', 'abrakadabra', 'x', 88.9)
    h['v'] = VectorHash([Hash('a', 21), Hash('z', Hash())])
    h['w'] = Hash('q.r', 77.7)
    # get iterator of ItemView...
    it = iter(h.items())
    k, v = next(it)
    assert k == 'a'
    assert v == Hash('b', 12)
    assert isinstance(v, Hash)
    # iterate over this Hash...
    jt = iter(v.items())
    k1, v1 = next(jt)
    assert k1 == 'b'
    assert isinstance(v1, int)
    assert v1 == 12
    with pytest.raises(StopIteration):
        next(jt)
    k, v = next(it)
    assert k == 'c'
    assert v == Hash('d', 23)
    k, v = next(it)
    assert k == 'e'
    assert v == 65
    k, v = next(it)
    assert k == 'f'
    assert v == Hash('g.m', 'abrakadabra')
    k, v = next(it)
    assert k == 'x'
    assert v == 88.9
    k, v = next(it)
    assert k == 'v'
    assert isinstance(v, VectorHash)
    assert v == VectorHash([Hash('a', 21), Hash('z', Hash())])
    assert len(v) == 2
    # iterate over VectorHash...
    jt = iter(v)
    hsh = next(jt)
    assert isinstance(hsh, Hash)
    assert hsh == Hash('a', 21)
    hsh = next(jt)
    assert hsh == Hash('z', Hash())
    with pytest.raises(StopIteration):
        next(jt)
    k, v = next(it)
    assert k == 'w'
    assert v == Hash('q.r', 77.7)
    with pytest.raises(StopIteration):
        next(it)

    # testing middlelayer-like API
    h = Hash('a.b.c', 21, 'f.g', "some string", 'i', 31.4865)
    # set attributes
    h['a.b.c', ...] = {'a': 12, 'b': {'e': "Invalid argument"}}
    assert h['a.b.c'] == 21
    assert h['i'] == 31.4865
    # retrieve attributes. Attributes syntax is 'dict'-like
    # 'attrs' is a ref to Hash node's attributes
    attrs = h['a.b.c', ...]
    assert attrs['a'] == 12
    assert attrs['b']['e'] == "Invalid argument"
    assert h['a.b.c', ...]['b']['e'] == "Invalid argument"
    assert h['a.b.c', 'b']['e'] == "Invalid argument"
    # update attributes
    h['a.b.c', ...].update({'w': True, 'b': {'e': "Valid argument"}})
    h['a.b.c', 'a'] = 13
    assert attrs['a'] == 13
    assert attrs['b']['e'] == "Valid argument"
    assert attrs['w'] is True
    h['a.b.c', 'w'] = False
    assert attrs['w'] is False
    # The 'attr' is a reference
    attrs['param'] = 'changed'
    assert h['a.b.c', ...]['param'] == 'changed'
    # ... or ...
    assert h['a.b.c', 'param'] == 'changed'
    # store attribits as a string
    sattrs = str(attrs)
    # delete parent Hash
    del h
    # check that 'attrs' reference is still alive
    try:
        s = str(attrs)
    except BaseException:
        s = None
        del attrs
    assert s == sattrs

    # add new attribute via looping with 'iterall'
    h = Hash('a', 1, 'b', 2, 'c', 3, 'd', 4)
    h['a', ...] = {'a': 12}
    h['b', ...] = {'b': 22}
    h['c', ...] = {'c': 32}
    h['d', ...] = {'d': 42}
    for k, _, a in h.iterall():
        a['w'] = 3.1415956
    assert h['a', 'a'] == 12
    assert h['a', 'w'] == 3.1415956
    assert h['b', 'b'] == 22
    assert h['b', 'w'] == 3.1415956
    assert h['c', 'c'] == 32
    assert h['c', 'w'] == 3.1415956
    assert h['d', 'd'] == 42
    assert h['d', 'w'] == 3.1415956

    # test 'HashAttributes'
    h = Hash('a.b.c', 21, 'f.g', "some string", 'i', 31.4865)
    # assign Hash attributes (lhs) the py dict (rhs) ...
    h['a.b.c', ...] = {'a': 12, 'b': {'e': "Invalid argument"}}

    g = Hash('x.y.z', 99)
    # assign attributes (python dict)
    g['x.y.z', ...] = {'x': 55, 'b': {'w': True}}
    # define the attributes ref (not a copy!)
    gattrs = g['x.y.z', ...]
    gattrs['x'] = 77
    assert g['x.y.z', 'x'] == 77
    assert g['x.y.z', ...]['x'] == 77
    assert gattrs['x'] == 77
    assert gattrs['b']['w'] is True

    # "merge"
    g['x.y.z', ...].update(h['a.b.c', ...])

    assert g['x.y.z', 'x'] == 77
    assert gattrs['x'] == 77
    assert gattrs['b']['e'] == "Invalid argument"
    assert 'w' not in gattrs['b']
    assert gattrs['a'] == 12

    # "replace" : 'g' attributes are the same as in 'h'
    # assign attributes (HashAttributes -> C++ Hash::Attributes)
    g['x.y.z', ...] = h['a.b.c', ...]

    # gattrs still points to the same object as g['x.y.z', ...]
    assert 'x' not in gattrs
    assert 'x' not in g['x.y.z', ...]
    assert gattrs['a'] == 12
    assert gattrs['b']['e'] == "Invalid argument"

    del gattrs
    del g


def test_getsetVectorHash():
    vh = VectorHash()
    vh.append(Hash('b', 1))
    vh.append(Hash('b', 2))
    h = Hash('a', vh)
    h['a'].extend((Hash('c', 3), Hash('c', 4),))
    h['a'].append(Hash('d', 5))
    assert len(h['a']) == 5
    assert h['a'][2]['c'] == 3
    assert h['a[2].c'] == 3
    assert isinstance(h, Hash)
    assert not isinstance(h['a'], Hash)
    assert isinstance(h['a'], VectorHash)
    assert isinstance(h['a'][2], Hash)
    assert isinstance(h['a[2].c'], int)
    h['a'][1]['b'] = 222
    assert h['a[1].b'] == 222

    h2 = h  # h2 points to the same tree
    h2['a[3].c'] = 33
    assert h2['a[3].c'] == 33
    assert h['a[3].c'] == 33

    h3 = copy.copy(h)  # full copy
    h3['a[3].c'] = 77
    assert h3['a[3].c'] == 77
    assert h['a[3].c'] == 33
    assert h == h2
    # assert not h == h3)

    # Check implicitly conversion [hash, hash,...] to VectorHash
    h = Hash('a', [Hash('a', 1), Hash('b', 2), Hash('c', 3)])
    assert h['a[0].a'] == 1
    assert h['a[1].b'] == 2
    assert h['a[2].c'] == 3


def test_copy_and_VectorHash():
    arr = np.arange(20000, dtype=np.int16).reshape(100, 200)
    assert arr[0][0] == 0
    v = VectorHash([Hash('a', arr), Hash('b', 2)])

    # Shallow copy ...
    vc = copy.copy(v)
    assert not id(vc) == id(v)

    vd = copy.deepcopy(v)
    assert not id(vd) == id(v)
    # Change origin vector...
    arr[0][0] = 111
    assert v[0]['a'][0][0] == 111
    # Shallow copy has seen the change ...
    assert vc[0]['a'][0][0] == 111
    # But deep copy is unchanged ...
    assert vd[0]['a'][0][0] == 0


def test_getAs():
    # BOOL
    h = Hash("a", True)
    assert h.getType("a") == Types.BOOL
    # The following tests are not needed since all conversion methods:
    # "getValueAs", "getAs", "getAttributeAs" use the same internal
    # conversion function.

    # Test "getAs"
    assert h.getAs("a", Types.BOOL)
    assert h.getAs("a", Types.CHAR) == '1'
    assert h.getAs("a", Types.INT8) == 1
    assert h.getAs("a", Types.UINT8) == 1
    assert h.getAs("a", Types.INT16) == 1
    assert h.getAs("a", Types.UINT16) == 1
    assert h.getAs("a", Types.INT32) == 1
    assert h.getAs("a", Types.UINT32) == 1
    assert h.getAs("a", Types.INT64) == 1
    assert h.getAs("a", Types.UINT64) == 1
    assert h.getAs("a", Types.FLOAT) == 1.0
    assert h.getAs("a", Types.DOUBLE) == 1.0
    assert h.getAs("a", Types.COMPLEX_FLOAT) == (1 + 0j)
    assert h.getAs("a", Types.COMPLEX_DOUBLE) == (1 + 0j)

    assert h.getAs("a", Types.STRING) == "1"
    assert h.getAs("a", Types.VECTOR_BOOL) == [True]
    # Gives bytearray(b'\xd4'):
    # assert h.getAs("a", Types.VECTOR_CHAR) == bytearray(b'\x01')
    assert h.getAs("a", Types.VECTOR_INT8) == [1]
    assert h.getAs("a", Types.VECTOR_UINT8) == [1]
    assert h.getAs("a", Types.VECTOR_INT16) == [1]
    assert h.getAs("a", Types.VECTOR_UINT16) == [1]
    assert h.getAs("a", Types.VECTOR_INT32) == [1]
    assert h.getAs("a", Types.VECTOR_UINT32) == [1]
    assert h.getAs("a", Types.VECTOR_INT64) == [1]
    assert h.getAs("a", Types.VECTOR_UINT64) == [1]
    assert h.getAs("a", Types.VECTOR_FLOAT) == [1.0]
    assert h.getAs("a", Types.VECTOR_DOUBLE) == [1.0]
    assert h.getAs("a", Types.VECTOR_COMPLEX_FLOAT) == [(1 + 0j)]
    assert h.getAs("a", Types.VECTOR_COMPLEX_DOUBLE) == [(1 + 0j)]
    assert h.getAs("a", Types.VECTOR_STRING) == ['1']


def test_getAttributeAs():
    # Test "getAttributesAs"
    h = Hash("a", True)
    h.setAttribute("a", "a", True)
    assert h.getAttributeAs("a", "a", Types.STRING), "1"
    assert h.getAttributeAs("a", "a", Types.CHAR) == '1'
    assert h.getAttributeAs("a", "a", Types.INT32) == 1
    assert h.getAttributeAs("a", "a", Types.DOUBLE) == 1.0
    assert h.getAttributeAs("a", "a", Types.VECTOR_BOOL) == [True]
    # Gives bytearray(b'\xd4')
    # assert
    #    h.getAttributeAs("a", "a", Types.VECTOR_CHAR) == bytearray(b'\x01')
    assert h.getAttributeAs("a", "a", Types.VECTOR_INT8) == [1]
    assert h.getAttributeAs("a", "a", Types.VECTOR_UINT8) == [1]
    assert h.getAttributeAs("a", "a", Types.VECTOR_INT32) == [1]
    assert h.getAttributeAs("a", "a", Types.VECTOR_DOUBLE) == [1.0]
    assert h.getAttributeAs("a", "a", Types.VECTOR_COMPLEX_FLOAT) == [(1 + 0j)]
    assert h.getAttributeAs("a", "a", Types.VECTOR_COMPLEX_DOUBLE) == [
        (1 + 0j)]
    assert h.getAttributeAs("a", "a", Types.VECTOR_STRING), ['1']

    # DOUBLE
    h.setAttribute("a", "c", 1.23)
    assert h.getAttributes("a").getNode('c').getType() == Types.DOUBLE
    assert h.getAttributeType('a', 'c') == Types.DOUBLE

    h = Hash('a', 1)
    # Check that `getAttributes(...)` return reference internal
    attrs = h.getAttributes("a")
    assert len(attrs) == 0
    h.setAttribute('a', 'a', True)
    assert len(attrs) == 1

    # Check that attributes can be copied to another Hash tree
    g = Hash("Z.a.b.c", "value")
    g.setAttributes("Z.a.b.c", attrs)
    assert g.getAttributeAs("Z.a.b.c", "a", Types.STRING) == "1"
    assert g.getAttributeAs("Z.a.b.c", "a", Types.INT32) == 1
    assert g.getAttributeAs("Z.a.b.c", "a", Types.DOUBLE) == 1.0


def test_setAs():
    # Test "setAs"
    testTypes = (Types.INT8, Types.UINT8, Types.INT16, Types.UINT16,
                 Types.INT32, Types.UINT32, Types.INT64, Types.UINT64,
                 Types.FLOAT, Types.DOUBLE)

    h = Hash()

    for t in testTypes:
        h.setAs("a", 5, t)
        assert h.getType("a") == t, "Setting as type " + str(t) + " failed"
        assert h["a"], 5 == "Equality failed for type " + str(t)

    # Special values for BOOL
    h.setAs("a", 1, "BOOL")
    assert h.getType("a") == Types.BOOL
    assert h["a"] is True

    h.setAs("a", 1.0, "VECTOR_BOOL")
    assert h.getType("a") == Types.VECTOR_BOOL
    assert h["a"] == [True]

    # Other vectors ...

    # Small positive numbers ... < 128
    h.setAs("a", [127, 0, 77], Types.VECTOR_INT8)
    assert h.getType("a") == Types.VECTOR_INT8
    # byte array: 0x7f -> 127, 0x00 -> 0, 'M' -> 77
    assert h["a"] == [127, 0, 77]

    h.setAs("a", [1260, -21170, 0, 1], Types.VECTOR_INT16)
    assert h.getType("a") == Types.VECTOR_INT16
    assert h["a"] == [1260, -21170, 0, 1]

    h.setAs("a", [1260, -21170, 0, 1], Types.VECTOR_DOUBLE)
    assert h.getType("a") == Types.VECTOR_DOUBLE
    assert h["a"] == [1260.0, -21170.0, 0.0, 1.0]

    h.setAs("a", [1260, -21170, 0, 1], Types.VECTOR_COMPLEX_FLOAT)
    assert h.getType("a") == Types.VECTOR_COMPLEX_FLOAT
    assert h["a"] == [(1260 + 0j), (-21170 + 0j), 0j, (1 + 0j)]

    h.setAs("a", [1260, -21170, 0, 1], Types.VECTOR_STRING)
    assert h.getType("a") == Types.VECTOR_STRING
    assert h["a"] == ['1260', '-21170', '0', '1']


def test_merge():
    h1 = Hash("a", 1,
              "b", 2,
              "c.b[0].g", 3,
              "c.c[0].d", 4,
              "c.c[1]", Hash("a.b.c", 6),
              "d.e", 7)
    h1.set("f.g", 99)
    h1.set("h", -1)

    # Test "setAttribute"
    h1.setAttribute("a", "attrKey", "Just a number")
    h1.setAttribute("c.b", "attrKey2", 3)
    h1.setAttribute("c.b[0].g", "attrKey3", 4.)
    h1.setAttribute("f", "attrKey6", "buaah!")

    h1b = Hash(h1)
    h1c = Hash(h1)
    h1d = Hash(h1)

    h2 = Hash("a", 21,
              "b.c", 22,
              "c.b[0]", Hash("key", "value"),
              "c.b[1].d", 24,
              "e", 27,
              "f", Hash())
    h2.set("g.h.i", -88)
    h2.set("g.h.j", -188)
    h2.set("h.i", -199)
    h2.set("h.j", 200)
    h2.set("i[3]", Hash())
    h2.set("i[1].j", 200)
    h2.set("i[2]", Hash("k.l", 5.))
    h2.set("j", Hash("k", 5.))
    h2.setAttribute("a", "attrKey", "Really just a number")
    h2.setAttribute("e", "attrKey4", -1)
    h2.setAttribute("e", "attrKey5", -11.)
    h2.setAttribute("f", "attrKey7", 77)
    # attribute on new vector<Hash> node
    h2.setAttribute("i", "attrKey8", 123)
    h2.setAttribute("j", "attrKey9", 12.3)  # ... and new Hash node

    h1.merge(h2)  # HashMergePolicy.REPLACE_ATTRIBUTES is the default
    h1b.merge(h2, HashMergePolicy.MERGE_ATTRIBUTES)
    # same as h1d.merge(h2): only check similarity and once
    # attribute replacement below
    h1d += h2

    assert similar(h1, h1b)
    assert similar(h1, h1d)

    assert h1.has("a")
    assert h1.get("a") == 21  # new value
    #  Attribute kept, but value overwritten:
    assert h1.hasAttribute("a", "attrKey"), "Attribute on node not kept"
    assert h1.getAttribute("a", "attrKey") == "Really just a number"
    assert 1 == h1.getAttributes("a").size()

    assert h1b.hasAttribute("a", "attrKey")
    assert h1b.getAttribute("a", "attrKey") == "Really just a number"

    assert 1, h1b.getAttributes("a").size()

    assert h1.has("b")
    assert h1.isType("b", Types.HASH)  # switch to new type
    assert h1.has("b.c")  # ...and as Hash can hold a child

    # Attributes overwritten by nothing or kept
    assert h1.getAttributes("c.b").size() == 0

    assert h1b.getAttributes("c.b").size() == 1
    assert h1b.hasAttribute("c.b", "attrKey2")
    assert h1b.getAttribute("c.b", "attrKey2") == 3

    assert not h1.has("c.b.d")
    assert h1.has("c.b[0]")
    assert h1.has("c.b[1]")
    assert h1.get("c.b[1].d") == 24
    assert h1.has("c.c[0].d")
    assert h1.has("c.c[1].a.b.c")
    assert h1.has("d.e")
    assert h1.has("e")
    assert h1.has("g.h.i")
    assert h1.has("g.h.j")
    assert h1.has("h.i")
    assert h1.has("h.j")
    assert h1.has("i[1].j")
    assert h1.has("i[2].k.l")
    assert h1.has("i[3]")
    assert h1.has("j.k")

    # Just add attributes with leaf (identical for REPLACE or MERGE)
    assert h1.getAttributes("e").size() == 2
    assert h1.getAttribute("e", "attrKey4") == -1
    assert h1.getAttribute("e", "attrKey5") == -11.
    assert h1b.getAttributes("e").size() == 2
    assert h1b.getAttribute("e", "attrKey4") == -1
    assert h1b.getAttribute("e", "attrKey5") == -11.
    # Just add attributes for new Hash/vector<Hash>
    # (identical for REPLACE or MERGE)
    assert h1.getAttributes("i").size() == 1
    assert h1.getAttribute("i", "attrKey8") == 123
    assert h1.getAttributes("j").size() == 1
    assert h1.getAttribute("j", "attrKey9") == 12.3

    assert h1b.getAttributes("i").size() == 1
    assert h1b.getAttribute("i", "attrKey8") == 123
    assert h1b.getAttributes("j").size() == 1
    assert h1b.getAttribute("j", "attrKey9") == 12.3
    assert h1b.hasAttribute("c.b", "attrKey2")

    assert h1.has("f")
    assert h1.has("f.g")  # merging does not overwrite h1["f"] with empty Hash

    assert h1.getAttributes("f").size() == 1
    assert h1.getAttribute("f", "attrKey7") == 77
    # += is merge with REPLACE_ATTRIBUTES
    assert h1d.getAttributes("f").size() == 1
    assert h1d.getAttribute("f", "attrKey7") == 77
    # here is MERGE_ATTRIBUTES
    assert h1b.getAttributes("f").size() == 2
    assert h1b.getAttribute("f", "attrKey6") == "buaah!"
    assert h1b.getAttribute("f", "attrKey7") == 77
    # Now check the 'selectedPaths' feature (no extra test
    # for attribute merging needed):
    selectedPaths = ["a", "b.c", "g.h.i", "h.i", "i[2]", "i[5]"]
    # check that we tolerate to select path with invalid index

    h1c.merge(h2, HashMergePolicy.MERGE_ATTRIBUTES, selectedPaths)

    # Keep everything it had before merging:
    assert h1c.has("a")
    assert h1c.has("b")
    assert h1c.has("c.b[0].g")
    assert h1c.has("c.c[0].d")
    assert h1c.has("c.c[1].a.b.c")
    assert h1c.has("d.e")
    assert h1c.has("f.g")
    # The additionally selected ones from h2:
    assert h1c.has("b.c")
    assert h1c.has("g.h.i")
    assert h1c.has("h.i")
    assert h1c.has("i[0].k.l")  # only row 2 selected, which becomes row 0
    # But not the other ones from h2:
    assert not h1c.has("c.b[0].key")  # neither at old position of h2
    assert not h1c.has("c.b[2]")  # nor an extended vector<Hash> at all
    assert not h1c.has("e")
    # Take care that adding path "g.h.i" does not trigger
    # that other children of "g.h"
    # in h2 are taken as well:
    assert not h1c.has("g.h.j")
    assert not h1c.has("h.j")
    # Adding i[2] should not trigger to add children of i[1] nor i[3]]
    assert not h1c.has("i[1].j")
    assert not h1c.has("i[3]")

    # Some further small tests for so far untested cases with
    # selected paths...
    hashTarget = Hash("a.b", 1, "a.c", Hash(), "c", "so so!")
    hashSource = Hash("a.d", 8., "ha", 9)
    selectedPaths = ("a",)
    # trigger merging a.d - tuple should work as well
    hashTarget.merge(hashSource, HashMergePolicy.MERGE_ATTRIBUTES,
                     selectedPaths)
    assert hashTarget.has("a.d")

    hashTargetB = Hash("a[1].b", 1, "c", "Does not matter")
    hashTargetC = Hash(hashTargetB)
    hashTargetD = Hash(hashTargetB)
    hashSourceBCD = Hash("a[2]", Hash("a", 33, "c", 4.4), "ha", 9,
                         "c[1]", Hash("k", 5, "l", 6),
                         "c[2]", Hash("b", -3), "d[2].b", 66,
                         "e[1]", Hash("1", 1, "2", 2, "3", 3))
    selectedPaths = ["a",  # trigger merging full vector
                     # trigger selecting first HashVec item
                     # overwriting what was not a hashVec before, but
                     # only keep selected items
                     "c[1].l",  # for rows no key selection, => '.l' is ignored
                     "d",  # trigger adding full new vector
                     "e[1].2",  # here '.2' is ignored
                     ]
    hashTargetB.merge(hashSourceBCD, HashMergePolicy.MERGE_ATTRIBUTES,
                      selectedPaths)
    assert hashTargetB.has("a[0]")  # The empty one merged into it
    assert not hashTargetB.has("a[0].b")
    assert hashTargetB.has("a[1]")  # dito
    assert not hashTargetB.has("a[1].b")  # target table a got replaced
    assert hashTargetB.has("a[2].a")
    assert not hashTargetB.has("a[5]")
    assert hashTargetB.has("c[0]")
    assert hashTargetB.has("c[0].k")
    assert hashTargetB.has("c[0].l")
    assert hashTargetB.has("d[2].b")
    assert not hashTargetB.has("d[3]")
    assert hashTargetB.has("e[0].1")
    assert hashTargetB.has("e[0].2")
    assert hashTargetB.has("e[0].3")

    selectedPaths = ("a[0]",
                     "a[2].b",  #
                     "c")  # trigger overwriting with complete vector
    hashTargetC.merge(hashSourceBCD, HashMergePolicy.MERGE_ATTRIBUTES,
                      selectedPaths)
    assert not hashTargetC.has("a[1].b")  # all table rows are overwritten
    assert hashTargetC.has("a[1].a")
    assert hashTargetC.has("a[1].c")
    assert not hashTargetC.has("a[2]")
    assert hashTargetC.has("c[1].k")
    assert hashTargetC.has("c[1].l")
    assert hashTargetC.has("c[2].b")
    assert not hashTargetC.has("c[3]")

    # Now select only invalid indices - nothing should be added
    selectedPaths = ["a[10]",  # to existing vector
                     "c[10]",  # where there was another node
                     "d[10]",  # where there was no node at all
                     "ha[0]"]  # for leaves, all indices are invalid
    copyD = Hash(hashTargetD)
    # also test use of keyword parameter:
    hashTargetD.merge(hashSourceBCD, selectedPaths=selectedPaths)
    assert similar(copyD, hashTargetD)

    h = Hash('a[0].a.b.c', 1, 'a[1].a.b.d', 2)
    g = Hash('a[0].x.y.w', 77, 'a[0].a.c', 33, 'a[2].abc', 12)
    h += g
    assert not h.has('a[0].a.b.c')  # overwritten
    assert h['a[0].x.y.w'] == 77
    assert h['a[0].a.c'] == 33
    assert not h.has('a[1].a.b.d')  # dito
    assert h['a[2].abc'] == 12


def test_subtract():
    h1 = Hash("a", 1,
              "b", 2,
              "c.b[0].g", 3,
              "c.c[0].d", 4,
              "c.c[1]", Hash("a.b.c", 6),
              "d.e", 7
              )

    h2 = Hash("a", 21,
              "b.c", 22,
              "c.b[0]", Hash("key", "value"),
              "c.b[1].d", 24,
              "e", 27
              )

    # Test "__iadd__"
    h1 += h2

    # Test "__isub__"
    h1 -= h2

    assert "a" not in h1
    assert h1["b"].empty()
    assert not h1.has("c.b[0].g")
    assert h1["c.c[0].d"] == 4
    assert h1["c.c[1].a.b.c"] == 6
    assert h1["d.e"] == 7

    h3 = Hash("a.b.c", 1,
              "a.b.d", 2,
              "b.c.d", 22,
              "c.a.b", 77,
              "c.c", "blabla"
              )
    h4 = Hash("a.b", Hash(), "c", Hash())
    h3 -= h4

    assert "a.b" in h3
    assert not h3["a"].empty()
    assert "c" in h3
    assert h3["b.c.d"] == 22


def test_erase():
    h = Hash('a[0].b[0].c', 1, 'b[0].c.d', 2, 'c.d[0].e', 3,
             'd.e', 4, 'e', 5, 'f.g.h.i.j.k', 6)
    assert 'a' in h
    assert 'f' in h
    assert 'b' in h
    assert 'c' in h
    assert 'd' in h

    paths = h.getPaths()
    for path in paths:
        if len(path) == 11:
            h.erasePath(path)
    assert 'a' not in h
    assert 'f' not in h
    assert 'b' in h
    assert 'c' in h
    assert 'd' in h

    paths = h.getPaths()
    for path in paths:
        if len(path) == 8:
            h.erasePath(path)
    assert 'a' not in h
    assert 'f' not in h
    assert 'b' not in h
    assert 'c' not in h
    assert len(h) == 2
    assert h['d.e'] == 4
    assert h['e'] == 5

    # now testing erase
    h1 = Hash('a', 1, 'b', 2, 'c.d', 31, 'e.f.g', 411, 'e.f.h', 412,
              'e.i', 42)

    assert len(h1) == 4

    # erase existing key on first level => size decreases
    assert h1.erase('a')
    assert 'a' not in h1
    assert len(h1) == 3

    # non-existing key - return false and keep size:
    assert not h1.erase('a')
    assert len(h1) == 3

    # 'c.d': composite key without siblings
    assert h1.erase('c.d')
    assert 'c.d' not in h1
    assert 'c' in h1
    assert len(h1) == 3  # 'c' still in!

    # 'e.f': composite key with two children and a sibling
    assert h1.erase('e.f')
    assert 'e.f.g' not in h1
    assert 'e.f.h' not in h1
    assert 'e.f' not in h1
    assert 'e' in h1  # stays
    assert len(h1) == 3


def test_fullEqual():
    h = Hash('a', 1, 'b', 2.2, 'c.d', "text",
             'e.f.g', VectorHash())
    h.setAttribute('a', "attr", 42)

    # A copy fully equals:
    h1 = copy.copy(h)
    assert fullyEqual(h1, h), "h1: " + str(h1)
    assert fullyEqual(h1, h, False), "h1: " + str(h1)

    # A changed value leads to being different
    h2 = copy.copy(h)
    h2["c.d"] = "TEXT"
    assert not fullyEqual(h2, h), "h2: " + str(h2)
    assert not fullyEqual(h2, h, False), "h2: " + str(h2)

    # An added value leads to being different
    h3 = copy.copy(h)
    h3["newKey"] = -11
    assert not fullyEqual(h3, h), "h3: " + str(h3)
    assert not fullyEqual(h3, h, False), "h3: " + str(h3)

    # A changed attribute leads to being different
    h4 = copy.copy(h)
    h4.setAttribute("a", "attr", -42)
    assert not fullyEqual(h4, h), "h4: " + str(h4)
    assert not fullyEqual(h4, h, False), "h4: " + str(h4)

    # An added attribute leads to being different
    h5 = copy.copy(h)
    h5.setAttribute("a", "newAttr", 0)
    assert not fullyEqual(h5, h), "h5: " + str(h5)
    assert not fullyEqual(h5, h, False), "h5: " + str(h5)

    # A changed order of keys leads to being different,
    # except if flag to ignore order is given
    h6 = copy.copy(h)
    bValue = h6["b"]
    h6.erase("b")
    h6["b"] = bValue
    assert not fullyEqual(h6, h), "h6: " + str(h6)
    assert fullyEqual(h6, h, False), "h6: " + str(h6)

    # A changed order of attribute keys leads to being different,
    # except if flag to ignore order is given
    h1.setAttribute("a", "attr2", 43)  # 2nd attribute
    h7 = copy.copy(h1)
    attr2Value = h7.getAttribute("a", "attr")
    h7.getAttributes("a").erase("attr")  # erase 1st attribute
    # re-add after previous 2nd
    h7.setAttribute("a", "attr", attr2Value)
    assert not fullyEqual(h7, h1), "h7: " + str(h7)
    assert fullyEqual(h7, h1, False), "h7: " + str(h7)


def test_ndarray():
    arr = np.array([[1, 2, 3, 4, 5], [6, 7, 8, 9, 10],
                    [11, 12, 13, 14, 15]])
    # The array owner is Python
    assert arr.flags.owndata
    # Put array into Hash ...
    h = Hash('a', arr)
    # The array owner inside the Hash is not a C++ code ...
    assert not h['a'].flags.owndata
    # The other attributes are not changed ...
    assert h['a'].dtype == np.dtype('int64')
    assert h['a'][0][0] == 1
    assert h['a'][0][3] == 4
    assert h['a'][2][1] == 12
    assert h['a'].size == 15
    assert h['a'].itemsize == 8
    assert h['a'].nbytes == 120
    assert h['a'].ndim == 2
    assert h['a'].shape == (3, 5)
    assert h['a'].strides == (40, 8)
    assert h['a'].flags.c_contiguous
    assert h['a'].flags.carray
    assert h['a'].flags.aligned
    assert h['a'].flags.writeable
    assert isinstance(h['a'], np.ndarray)

    # copy.copy will NOT copy the underlying C++ NDArray
    h2 = copy.copy(h)

    assert isinstance(h2['a'], np.ndarray)
    assert h2['a'][0][0] == 1

    h['a'][0][0] = 255

    assert h['a'][0][0] == 255
    assert h2['a'][0][0] == 255
    assert arr[0][0] == 255

    # Check 'deepcopy': it copies even underlying NDArray
    g = copy.deepcopy(h)
    assert g['a'][0][0] == 255

    h['a'][0][0] = 1

    assert h['a'][0][0] == 1
    assert g['a'][0][0] == 255

    # Check fortran order array:
    # Since the C++ NDArray only knows C order, it has to be copied
    arr = np.array([[1, 2, 3], [4, 5, 6]], order='F')
    assert arr.flags.f_contiguous
    assert not arr.flags.c_contiguous

    h['b'] = arr  # copies!
    h_arr = h['b']
    assert not h_arr.flags.f_contiguous
    assert h_arr.flags.c_contiguous
    assert h_arr[0][0] == 1
    assert h_arr[0][1] == 2
    assert h_arr[0][2] == 3
    assert h_arr[1][0] == 4
    assert h_arr[1][1] == 5
    assert h_arr[1][2] == 6
    arr[0][0] == 11
    assert h_arr[0][0] == 1  # untouched

    arr = np.arange(20000, dtype=np.int16).reshape(100, 200)
    arr.flags.writeable = False
    h = Hash()
    h['a'] = arr
    assert np.all(h['a'] == arr) is np.True_
    # Can't insert non-writable arrays into Hash
    # with pytest.raises(ValueError):
    #     h['a'] = arr



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

from karabo.bound import (
    ALARM_ELEMENT, BOOL_ELEMENT, DOUBLE_ELEMENT, FLOAT_ELEMENT,
    IMAGEDATA_ELEMENT, INT32_ELEMENT, INT64_ELEMENT, KARABO_CLASSINFO,
    KARABO_CONFIGURATION_BASE_CLASS, NDARRAY_ELEMENT, NODE_ELEMENT,
    OVERWRITE_ELEMENT, SLOT_ELEMENT, STATE_ELEMENT, STRING_ELEMENT,
    TABLE_ELEMENT, UINT32_ELEMENT, UINT64_ELEMENT, VECTOR_BOOL_ELEMENT,
    VECTOR_DOUBLE_ELEMENT, VECTOR_INT32_ELEMENT, VECTOR_STRING_ELEMENT,
    AlarmCondition, ArchivePolicy, Hash, MetricPrefix, Schema, Unit)
from karabo.common.states import State


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("Shape", "1.0")
class Shape:
    def __init__(self, configuration):
        self.configuration = configuration

    @staticmethod
    def expectedParameters(expected):
        (
            BOOL_ELEMENT(expected).key("shadowEnabled")
            .description("Shadow enabled")
            .displayedName("Shadow")
            .assignmentOptional().defaultValue(False)
            .init()
            .commit(),
        )

    def getConfiguration(self):
        return self.configuration

    def draw(self):
        pass


@KARABO_CLASSINFO("Circle", "1.0")
class Circle(Shape):
    def __init__(self, configuration):
        super().__init__(configuration)

    @staticmethod
    def expectedParameters(expected):
        (
            DOUBLE_ELEMENT(expected).key("radius").alias(1)
            .description("The radius of the circle")
            .displayedName("Radius")
            .minExc(0).maxExc(100)
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(10)
            .init()
            .commit(),

            STATE_ELEMENT(expected).key("state")
            .commit(),

            STATE_ELEMENT(expected).key("stateN")
            .defaultValue(State.NORMAL)
            .commit(),

            STATE_ELEMENT(expected).key("stateE")
            .initialValue(State.ERROR)
            .commit(),

            ALARM_ELEMENT(expected).key("alarm")
            .commit(),

            ALARM_ELEMENT(expected).key("alarmW")
            .initialValue(AlarmCondition.WARN)
            .commit(),

            ALARM_ELEMENT(expected).key("alarmA")
            .defaultValue(AlarmCondition.ALARM)
            .commit(),

            STRING_ELEMENT(expected).key("status")
            .readOnly()
            .commit()
        )

    def draw(self):
        return self.__class__.__name__


'''
Editable Circle
'''


@KARABO_CLASSINFO("EditableCircle", "1.0")
class EditableCircle(Circle):
    def __init__(self, configuration):
        super().__init__(configuration)

    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("radius")
            .setNowReconfigurable()
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("state")
            .setNewOptions(State.INIT, State.ERROR, State.NORMAL)
            .setNewDefaultValue(State.INIT)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("status")
            .setNewOptions("a,b,c")
            .setNewDefaultValue("a")
            .commit()
        )

    def draw(self):
        return self.__class__.__name__


'''
Rectangle
'''


@KARABO_CLASSINFO("Rectangle", "1.0")
class Rectangle(Shape):
    def __init__(self, configuration):
        super().__init__(configuration)
        # print "Rectangle.__init__"

    @staticmethod
    def expectedParameters(expected):
        (
            DOUBLE_ELEMENT(expected).key("a")
            .alias(1)
            .description("Length of a")
            .displayedName("A")
            .minExc(0).maxExc(100)
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(10)
            .init()
            .commit(),

            DOUBLE_ELEMENT(expected).key("b").alias(1)
            .description("Length of b")
            .displayedName("B")
            .minExc(0).maxExc(100)
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(10)
            .init()
            .commit(),
        )

    def draw(self):
        return self.__class__.__name


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("GraphicsRenderer", "1.0")
class GraphicsRenderer:
    def __init__(self, input):
        selected = input["shapes"]
        shape = Shape.create(selected, input[selected])
        if "Circle" in input:
            assert shape.draw() == "Circle"

    @staticmethod
    def expectedParameters(expected):
        (
            BOOL_ELEMENT(expected).key("antiAlias")
            .tags("prop")
            .displayedName("Use Anti-Aliasing")
            .description("You may switch of for speed")
            .assignmentOptional().defaultValue(True)
            .init()
            .expertAccess()
            .commit(),

            STRING_ELEMENT(expected).key("color")
            .tags("prop")
            .displayedName("Color")
            .options("red,green,blue,orange,black")
            .description("The default color for any shape")
            .assignmentOptional().defaultValue("red")
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected).key("bold")
            .tags("prop")
            .displayedName("Bold")
            .description("Toggles bold painting")
            .assignmentOptional().defaultValue(False)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected).key("shapes")
            .description("Some shapes")
            .displayedName("Shapes")
            .options("Circle, EditableCircle, Rectangle")
            .assignmentOptional().defaultValue("Rectangle")
            .commit(),

            NODE_ELEMENT(expected)
            .key("Circle")
            .appendParametersOfConfigurableClass(Shape, "Circle")
            .commit(),

            NODE_ELEMENT(expected)
            .key("EditableCircle")
            .appendParametersOfConfigurableClass(Shape, "EditableCircle")
            .commit(),

            NODE_ELEMENT(expected)
            .key("Rectangle")
            .appendParametersOfConfigurableClass(Shape, "Rectangle")
            .commit(),
        )


@KARABO_CLASSINFO("GraphicsRenderer1", "1.0")
class GraphicsRenderer1:
    @staticmethod
    def expectedParameters(expected):
        (
            BOOL_ELEMENT(expected).key("antiAlias")
            .tags("prop")
            .displayedName("Use Anti-Aliasing")
            .description("You may switch of for speed")
            .assignmentOptional().defaultValue(True)
            .init()
            .expertAccess()
            .commit(),

            STRING_ELEMENT(expected).key("color")
            .tags("prop")
            .displayedName("Color")
            .description("The default color for any shape")
            .assignmentOptional().defaultValue("red")
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected).key("bold")
            .tags("prop")
            .displayedName("Bold")
            .description("Toggles bold painting")
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected).key("shapes")
            .assignmentOptional().defaultValue("circle")
            .commit(),

            NODE_ELEMENT(expected).key("circle")
            .tags("shape")
            .displayedName("Circle")
            .description("A circle")
            .appendParametersOf(Circle)
            .commit(),

            NODE_ELEMENT(expected).key("rectangle")
            .tags("shape")
            .displayedName("Rectangle")
            .description("A rectangle")
            .commit(),

            DOUBLE_ELEMENT(expected).key("rectangle.b")
            .description("Rectangle side - b")
            .displayedName("Side B")
            .assignmentOptional().defaultValue(10)
            .init()
            .commit(),

            DOUBLE_ELEMENT(expected).key("rectangle.c")
            .description("Rectangle side - c")
            .displayedName("Side C")
            .assignmentOptional().defaultValue(10)
            .init()
            .commit(),

            NODE_ELEMENT(expected).key("triangle")
            .displayedName("triangle")
            .description("A triangle (Node element containing no "
                         "other elements)")
            .commit(),
        )


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("TestStruct1", "1.0")
class TestStruct1:
    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected).key("exampleKey1")
            .tags("hardware, poll")
            .displayedName("Example key 1")
            .description("Example key 1 description")
            .options("Radio,Air Condition,Navigation", ",")
            .assignmentOptional().defaultValue("Navigation")
            .reconfigurable()
            .operatorAccess()
            .commit(),

            INT32_ELEMENT(expected).key("exampleKey2")
            .alias(10)
            .tags("hardware, poll")
            .displayedName("Example key 2")
            .description("Example key 2 description")
            .options("5, 25, 10")
            .minInc(5).maxInc(25)
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(10)
            .init()
            .operatorAccess()
            .commit(),

            UINT32_ELEMENT(expected).key("exampleKey3")
            .alias(5.5)
            .tags("hardware, set")
            .displayedName("Example key 3")
            .description("Example key 3 description")
            .allowedStates(State.STARTED, State.STOPPED, State.NORMAL)
            .minExc(10).maxExc(20)
            .assignmentMandatory()
            .expertAccess()
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(expected).key("exampleKey4")
            .alias("exampleAlias4")
            .tags("software")
            .displayedName("Example key 4")
            .description("Example key 4 description")
            .options("1.11     -2.22 5.55")
            .minExc(-2.22).maxExc(5.55)
            .assignmentInternal().noDefaultValue()
            .expertAccess()
            .commit(),

            INT64_ELEMENT(expected).key("exampleKey5")
            .alias("exampleAlias5")
            .tags("h/w; d.m.y", ";")
            .displayedName("Example key 5")
            .description("Example key 5 description")
            .setSpecialDisplayType("Int64DisplayType")
            .readOnly().initialValue(1442244)
            .archivePolicy(ArchivePolicy.EVERY_EVENT)
            .commit(),

            INT64_ELEMENT(expected).key("exampleKey5b")
            .readOnly().defaultValue(42)
            .commit(),

            DOUBLE_ELEMENT(expected).key("exampleKey6")
            .alias([0x00123456, 0x0000aacc])
            .displayedName("Example key 6")
            .description("Example key 6 description")
            .readOnly().initialValue(1.11)
            .commit(),

            VECTOR_INT32_ELEMENT(expected).key("exampleKey7")
            .displayedName("Example key 7")
            .allowedStates(State.STARTED, State.NORMAL)
            .readOnly().initialValue([1, 2, 3])
            .commit(),

            VECTOR_INT32_ELEMENT(expected).key("exampleKey7b")
            .readOnly().initialValue([11, 22, 33])
            .commit(),

            VECTOR_DOUBLE_ELEMENT(expected).key("exampleKey8")
            .readOnly().initialValue([1.1, 2.2, 3.3])
            .archivePolicy(ArchivePolicy.NO_ARCHIVING)
            .commit(),

            VECTOR_STRING_ELEMENT(expected).key("exampleKey9")
            .setSpecialDisplayType("TestDisplayType")
            .readOnly().initialValue(["Hallo", "World"])
            .commit(),

            VECTOR_INT32_ELEMENT(expected).key("vectInt")
            .readOnly()
            .commit(),

            VECTOR_INT32_ELEMENT(expected).key("exampleKey10")
            .displayedName("Example key 10")
            .minSize(2)
            .maxSize(7)
            .assignmentOptional().defaultValueFromString("10, 20, 30")
            .reconfigurable()
            .commit(),

            VECTOR_INT32_ELEMENT(expected).key("exampleKey11")
            .displayedName("Example key 11")
            .assignmentOptional().defaultValue([10, 20, 30])
            .observerAccess()
            .reconfigurable()
            .commit(),

            VECTOR_DOUBLE_ELEMENT(expected).key("exampleKey12")
            .assignmentOptional().defaultValueFromString("1.1, -2.2, 3.3")
            .reconfigurable()
            .commit(),

            VECTOR_STRING_ELEMENT(expected).key("exampleKey14")
            .assignmentOptional().defaultValue(["Hallo", "World", "Test"])
            .reconfigurable()
            .commit(),

            VECTOR_STRING_ELEMENT(expected).key("exampleKey15")
            .assignmentOptional()
            .defaultValueFromString("word1, word2, test")
            .reconfigurable()
            .commit(),

            NDARRAY_ELEMENT(expected).key("exampleKey16")
            .dtype("UINT32")
            .shape([2, 5, 0])
            .commit(),

            INT32_ELEMENT(expected).key("exampleIntKey")
            .assignmentOptional().defaultValueFromString("20")
            .reconfigurable()
            .commit(),

            UINT32_ELEMENT(expected).key("exampleBitsKey1")
            .tags("hardware")
            .displayedName("Example bits key 1")
            .description("Example bits key 1 description")
            .reconfigurable()
            .bin()
            .assignmentOptional().defaultValue(0xdeadbeef)
            .commit(),

            UINT64_ELEMENT(expected).key("exampleBitsKey2")
            .tags("hardware")
            .displayedName("Example bits key 2")
            .description("Example bits key 2 description")
            .reconfigurable()
            .bin("10:In Error, 21:Busy, 35:HV On, 55:Crate On")
            .assignmentOptional().defaultValue(0xdeadbeefdeadface)
            .commit(),

            UINT32_ELEMENT(expected).key("exampleBitsKey3")
            .tags("hardware")
            .displayedName("Example bits key 3")
            .description("Example bits key 3 description")
            .reconfigurable()
            .oct()
            .assignmentOptional().defaultValue(0xbeefface)
            .commit(),

            UINT32_ELEMENT(expected).key("exampleBitsKey4")
            .tags("hardware")
            .displayedName("Example bits key 4")
            .description("Example bits key 4 description")
            .reconfigurable()
            .hex()
            .assignmentOptional().defaultValue(0xbeefface)
            .commit(),

            STRING_ELEMENT(expected).key("testPath")
            .alias(5)
            .displayedName("Filename")
            .options("file1, file2, karabo.log")
            .assignmentOptional().defaultValue("karabo.log")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected).key("testPath2")
            .readOnly()
            .commit(),

            STRING_ELEMENT(expected).key("testPath3")
            .assignmentMandatory()
            .commit(),

            SLOT_ELEMENT(expected).key("slotTest")
            .displayedName("Reset")
            .description("Test slot element")
            .allowedStates(State.STARTED, State.STOPPED, State.NORMAL)
            .commit(),

            IMAGEDATA_ELEMENT(expected).key("myImageElement")
            .displayedName("myImage")
            .description("Image Element")
            .setDimensions("100, 200")
            .operatorAccess()
            .commit(),
        )
        row = Schema()
        INT32_ELEMENT(row).key("int").readOnly().initialValue(0).commit()
        (
            TABLE_ELEMENT(expected).key("tableI")
            .setColumns(row)
            .readOnly()
            .initialValue([Hash("int", 2)])
            .commit(),

            TABLE_ELEMENT(expected).key("tableD")
            .setColumns(row)
            .readOnly()
            .defaultValue([Hash("int", 3)])
            .commit(),

            NODE_ELEMENT(expected).key("myNode")
            .displayedName("myNode")
            .description("WidgetNode Test")
            .setSpecialDisplayType("WidgetNode")
            .commit(),
        )


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("SomeClassId", "1.0")
class SomeClass:
    def __init__(self, configuration):
        super().__init__()

    @staticmethod
    def expectedParameters(expected):
        (
            INT32_ELEMENT(expected).key("x").alias(10)
            .tags("IK,BH")
            .displayedName("Xkey")
            .description("Example of X key description")
            .options("5, 25, 10")
            .minInc(5).maxInc(25).unit(Unit.AMPERE).metricPrefix(
                MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(5)
            .init()
            .expertAccess()
            .commit(),

            INT32_ELEMENT(expected).key("y").alias('bla')
            .tags("CY")
            .displayedName("Ykey")
            .description("Example of Y key description")
            .options("5, 25, 10")
            .minExc(0).maxExc(29).unit(Unit.METER).metricPrefix(
                MetricPrefix.CENTI)
            .assignmentOptional().defaultValue(10)
            .init()
            .commit(),

            DOUBLE_ELEMENT(expected).key("a")
            .readOnly().initialValue(1.11)
            .archivePolicy(ArchivePolicy.NO_ARCHIVING)
            .commit(),

            VECTOR_INT32_ELEMENT(expected).key("somelist")
            .displayedName("Ykey")
            .description("Example of Y key description")
            .reconfigurable()
            .assignmentOptional().defaultValue([])
            .commit(),

            IMAGEDATA_ELEMENT(expected).key("myImageElement")
            .displayedName("myImage")
            .description("Image Element")
            .setDimensions([110, 210])
            .operatorAccess()
            .commit(),
        )


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("Base", "1.0")
class Base:
    def __init__(self, configuration):
        super().__init__()


@KARABO_CLASSINFO("P1", "1.0")
class P1(Base):
    def __init__(self, configuration):
        super().__init__(configuration)

    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected).key("a")
            .description("a").displayedName("a")
            .assignmentOptional().defaultValue("a value")
            .tags("CY,CY,NC,JS,KW,NC")
            .commit(),

            STRING_ELEMENT(expected).key("b")
            .tags("BH,CY")
            .displayedName("Example key 1")
            .description("Example key 1 description")
            .options("Radio,Air Condition,Navigation", ",")
            .assignmentOptional().defaultValue("Air Condition")
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(expected).key("c").alias(10)
            .tags("BH")
            .displayedName("Example key 2")
            .description("Example key 2 description")
            .options("5, 25, 10")
            .minInc(5).maxInc(25).unit(Unit.METER).metricPrefix(
                MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(5)
            .init()
            .commit(),

            UINT32_ELEMENT(expected).key("d").alias(5.5)
            .tags("CY,JS")
            .displayedName("Example key 3")
            .description("Example key 3 description")
            .allowedStates(State.STARTED, State.STOPPED,
                           State.NORMAL)  # TODO check
            .minExc(10).maxExc(20).assignmentOptional().defaultValue(11)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(expected).key("e").alias("exampleAlias4")
            .tags("DB,NC,CY")
            .displayedName("Example key 4")
            .description("Example key 4 description")
            .options("1.1100000, -2.22, 5.55")
            .assignmentOptional()
            .defaultValue(1.11)
            .commit(),

            INT64_ELEMENT(expected).key("f").alias("exampleAlias5")
            .tags("LM,DB")
            .displayedName("Example key 5")
            .description("Example key 5 description")
            .assignmentOptional()
            .defaultValue(5)
            .commit(),
        )


@KARABO_CLASSINFO("P2", "1.0")
class P2(Base):
    def __init__(self, configuration):
        super().__init__(configuration)

    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected).key("x")
            .description("x")
            .displayedName("x")
            .assignmentOptional().defaultValue("a value")
            .tags("LM,BH")
            .commit(),

            STRING_ELEMENT(expected).key("y")
            .tags("CY")
            .displayedName("Example key 1")
            .description("Example key 1 description")
            .options("Radio,Air Condition,Navigation", ",")
            .assignmentOptional().defaultValue("Radio")
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(expected).key("z").alias(10)
            .tags("CY,LM,KW")
            .displayedName("Example key 2")
            .description("Example key 2 description")
            .options("5, 25, 10")
            .minInc(5).maxInc(25).unit(Unit.AMPERE).metricPrefix(
                MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(10)
            .init()
            .commit(),
        )


@KARABO_CLASSINFO("P3", "1.0")
class P3(Base):
    def __init__(self, configuration):
        super().__init__(configuration)

    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected).key("k")
            .description("k").displayedName("k")
            .assignmentOptional().defaultValue("k value")
            .tags("LM")
            .commit(),

            STRING_ELEMENT(expected).key("l")
            .tags("CY")
            .displayedName("l").description("l")
            .options("Radio,Air Condition,Navigation", ",")
            .assignmentOptional().defaultValue("Navigation")
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(expected).key("m").alias(10)
            .tags("CY,DB,JE,BP,MK,PG,BF")
            .displayedName("Example key 2")
            .description("Example key 2 description")
            .options("5, 25, 10")
            .minInc(5).maxInc(25).unit(Unit.METER).metricPrefix(
                MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(25)
            .init()
            .commit(),
        )


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("ArrayContainer", "1.0")
class ArrayContainer(Base):
    def __init__(self, configuration):
        super().__init__(configuration)

    @staticmethod
    def expectedParameters(expected):
        (
            NDARRAY_ELEMENT(expected).key("exampleKey16")
            .dtype("BOOL")
            .shape([2, 3])
            .unit(Unit.DEGREE_CELSIUS)
            .metricPrefix(MetricPrefix.CENTI)
            .commit(),

            NDARRAY_ELEMENT(expected).key("exampleKey17")
            .dtype("UINT32")
            .shape([2, 5, 0])
            .commit(),
            NDARRAY_ELEMENT(expected).key("exampleKey18")
            .dtype("FLOAT")
            .shape("3,2,1")
            .commit()
        )


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("GraphicsRenderer2", "1.0")
class GraphicsRenderer2:
    def __init__(self, configuration):
        super().__init__()

    @staticmethod
    def expectedParameters(expected):
        (
            BOOL_ELEMENT(expected).key("antiAlias")
            .tags("NC")
            .displayedName("Use Anti-Aliasing")
            .description("You may switch of for speed")
            .assignmentOptional().defaultValue(True)
            .init().expertAccess()
            .commit(),

            STRING_ELEMENT(expected).key("color")
            .tags("KW")
            .displayedName("Color")
            .description("The default color for any shape")
            .assignmentOptional().defaultValue("red")
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected).key("bold")
            .tags("LM")
            .displayedName("Bold").description("Toggles bold painting")
            .assignmentOptional().defaultValue(False)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected).key("shapes")
            .tags("DB")
            .assignmentOptional().defaultValue("rectangle")
            .commit(),

            NODE_ELEMENT(expected).key("circle")
            .tags("JS")
            .displayedName("Circle").description("A circle")
            .commit(),

            FLOAT_ELEMENT(expected).key("circle.radius")
            .description("The radius of the circle")
            .displayedName("Radius")
            .tags("NC,KW")
            .minExc(0).maxExc(100).unit(Unit.METER).metricPrefix(
                MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(10)
            .init().commit(),

            NODE_ELEMENT(expected).key("rectangle")
            .tags("BH, KW , CY")
            .displayedName("Rectangle").description("A rectangle")
            .commit(),

            FLOAT_ELEMENT(expected).key("rectangle.b")
            .tags("JS")
            .description("Rectangle side - b").displayedName("Side B")
            .assignmentOptional().defaultValue(10)
            .init().commit(),

            FLOAT_ELEMENT(expected).key("rectangle.c")
            .tags("LM,JS")
            .description("Rectangle side - c").displayedName("Side C")
            .assignmentOptional().defaultValue(10)
            .init()
            .commit(),

            NODE_ELEMENT(expected).key("triangle")
            .displayedName("triangle")
            .description("A triangle (Node element containing no "
                         "other elements)")
            .commit(),

            NODE_ELEMENT(expected).key("letter")
            .displayedName("Letter").description("Letter")
            .appendParametersOf(P1)
            .commit(),

        )


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("OtherSchemaElements", "1.0")
class OtherSchemaElements:
    def __init__(self, configuration):
        super().__init__()

    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected)
            .description("File name")
            .key("filename")
            .alias(5)
            .displayedName("Filename")
            .options("file1, file2, karabo.log")
            .assignmentOptional()
            .defaultValue("karabo.log")
            .reconfigurable()
            .commit(),

            VECTOR_INT32_ELEMENT(expected)
            .key("vecIntReconfig")
            .assignmentOptional()
            .defaultValue([10, 20, 30])
            .reconfigurable()
            .commit(),

            VECTOR_INT32_ELEMENT(expected)
            .key("vecIntReconfigStr")
            .assignmentOptional()
            .defaultValueFromString("11, 22, 33")
            .reconfigurable()
            .commit(),

            VECTOR_DOUBLE_ELEMENT(expected)
            .key("vecDoubleReconfigStr")
            .assignmentOptional()
            .defaultValueFromString("1.1, 2.2, 3.3")
            .reconfigurable()
            .commit(),

            VECTOR_BOOL_ELEMENT(expected)
            .key("vecBool")
            .tags("h/w; d.m.y", ";")
            .allowedStates(State.STARTED, State.STOPPED)
            .minSize(2)
            .maxSize(7)
            .assignmentMandatory()
            .commit(),
        )



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

import numpy as np

from karabo.bound import Dims, Encoding, Hash, ImageData, Rotation, Types
from karabo.testing.utils import compare_ndarray_data_ptrs


def test_imagedata_buildUp():
    # Create numpy array ...
    arr = np.array([[1, 2], [3, 4], [5, 6]])
    # By default: 64 bits
    assert arr.dtype == np.dtype('int64')
    # check shape...
    assert arr.shape == (3, 2)
    # Python is an owner
    assert arr.flags.owndata is True
    # Create default image data from 'arr'
    img = ImageData(arr)
    # image data is not an owner
    assert img.getData().flags.owndata is False
    assert np.all(arr == img.getData())
    # check binning
    assert img.getBinning() == (1, 1)
    # check deduced bits per pixel ...
    assert img.getBitsPerPixel() == 64
    # check deduced dimensions ...
    assert img.getDimensions() == (3, 2)
    # check deduced encoding ...
    assert img.getEncoding() == Encoding.GRAY
    # check flip in X scale
    assert img.getFlipX() is False
    # ... and Y scale
    assert img.getFlipY() is False
    # Get ROI offesets ...
    assert img.getROIOffsets() == (0, 0)
    # No rotation ... (0 grad)
    assert img.getRotation() == Rotation.ROT_0
    # check type ...
    assert img.getType() == 16
    # is indexable?
    assert img.isIndexable() is True
    # create image with defined dimensions and encoding...
    # one-dimensional array ...
    arr = np.array([1, 2, 3, 4, 5, 6])
    # define required shape...
    dims = Dims(3, 2)
    img = ImageData(arr, dims, Encoding.RGBA)
    assert img.getEncoding() == Encoding.RGBA
    img.setEncoding(Encoding.GRAY)
    assert img.getEncoding() == Encoding.GRAY
    img.setBitsPerPixel(12)
    assert img.getBitsPerPixel() == 12
    # check shape of image
    assert img.getData().shape == (3, 2)
    # rotate 180 grad
    img.setRotation(Rotation.ROT_180)
    assert img.getRotation() == Rotation.ROT_180

    # Check that C-order is guaranteed in C++ NDArray
    # The original order is not kept and returned from C++ as C-style.

    # Input array is F_CONTIGUOUS array
    x = np.asarray(np.arange(20000, dtype=np.int16).reshape(100, 200),
                   order='F')
    assert x.flags.owndata is True
    assert x.flags.c_contiguous is False
    assert x.flags.f_contiguous is True
    ix = ImageData(x)
    # array inside ImageData is kept as C++ NDArray in c-order
    assert not ix.getData().flags.owndata
    assert ix.getData().flags.c_contiguous is True
    assert ix.getData().flags.f_contiguous is False
    assert compare_ndarray_data_ptrs(x, ix.getData()) is False

    # Input array is C_CONTIGUOUS array
    y = np.asarray(np.arange(20000, dtype=np.int16).reshape(100, 200),
                   order='C', copy=True)
    assert compare_ndarray_data_ptrs(x, y) is False
    assert y.flags.owndata is True
    assert y.flags.c_contiguous is True
    assert y.flags.f_contiguous is False
    assert np.all(x == y) is np.True_
    assert compare_ndarray_data_ptrs(x, y) is False
    iy = ImageData(y)
    assert iy.getData().flags.c_contiguous is True
    assert iy.getData().flags.f_contiguous is False
    assert compare_ndarray_data_ptrs(y, iy.getData()) is True

    assert np.all(ix.getData() == iy.getData())
    assert compare_ndarray_data_ptrs(ix.getData(), iy.getData()) is False

    # getDataCopy: the new python object will be created
    nx = ix.getDataCopy()
    assert nx.flags.owndata is True
    assert nx.flags.c_contiguous is True
    assert nx.flags.f_contiguous is False
    ny = iy.getDataCopy()
    assert ny.flags.owndata is True
    assert ny.flags.c_contiguous is True
    assert ny.flags.f_contiguous is False

    assert np.all(nx == ny) == np.True_
    assert np.all(ix.getData() == nx) == np.True_

    # getData/setData
    assert x.flags.owndata is True
    assert x.flags.f_contiguous is True
    img = ImageData()
    img.setData(x)
    assert img.getData().flags.owndata is False
    assert img.getData().flags.c_contiguous is True
    assert np.all(img.getData() == ix.getData()) == np.True_

    # setDataCopy
    img = ImageData()  # empty ImageData object
    img.setDataCopy(ix.getData())
    assert img.getData().flags.owndata is False
    assert img.getData().flags.c_contiguous is True
    assert img.getData().flags.f_contiguous is False
    assert np.all(img.getData() == ix.getData()) == np.True_

    # getType/setType
    assert img.getType() == Types.INT16
    img.setType(Types.UINT16)
    assert img.getType() == Types.UINT16

    # getDimensions/setDimensions
    assert img.getDimensions() == (100, 200)
    img.setDimensions([200, 100])
    assert img.getDimensions() == (200, 100)

    # getROIOffsets/setROIOffsets
    assert img.getROIOffsets() == (0,)
    img.setROIOffsets((30, 40))
    assert img.getROIOffsets() == (30, 40)

    # getBinning/setBinning
    binning = Dims(3, 8)
    img.setBinning(binning)  # or img,setBinning((3, 8))
    assert img.getBinning() == (3, 8)

    # setFlipX/Y
    img.setFlipY(True)
    img.setFlipX(False)
    assert img.getFlipX() is False
    assert img.getFlipY() is True


def test_imagedata_hash():
    # Create numpy array
    arr = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    # Python is an owner of data
    assert arr.flags.owndata is True
    # create image data object...
    # NOTE: Dims defines shape
    img = ImageData(arr, Dims(3, 4), Encoding.RGBA, 32)
    # delete original array
    del arr
    # Store image data into Hash
    h = Hash('image', img, 'a', 12)
    assert h['image'].getData().flags.owndata is False
    assert isinstance(h['image'], ImageData) is True
    assert h['image'].getData().shape == (3, 4)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
# To change this template, choose Tools | Templates
# and open the template in the editor.

from karabind import AssemblyRules, Schema, Validator


class Configurator:
    """Provides factorized configuration

    Configurator is the singleton class that keeps methods for registration and
    creation other classes that are configurable classes
    """
    _instance = None
    registry = {}

    @staticmethod
    def registerAsBaseClass(theClass):
        """Register a class as a base class in the configurator

        :param theClass: to be registered
        """
        if theClass.__classid__ not in Configurator.registry:
            theClass.__base_classid__ = theClass.__classid__
            theClass.__bases_classid__ = (theClass.__classid__,)
            Configurator.registry[theClass.__classid__] = {}
        # self-registering
        shrt = theClass.__classid__
        Configurator.registry[shrt][shrt] = theClass

    def __init__(self, classid):
        """
        The argument to the constructor may be the classid of a configurable
        class or configurable class itself:
        Example:
                c = Configurator(ConfigurableClass)
        or
                c = Configurator("ConfigurableClassId")
        """
        if isinstance(classid, type):
            classid = classid.__classid__
        if not isinstance(classid, str):
            raise TypeError("The argument type '{}' is not allowed. "
                            "Must be a class or a str.".format(type(classid)))
        if classid not in Configurator.registry:
            raise AttributeError(
                "Argument is not a class or classid of registered base class")
        self.baseRegistry = Configurator.registry[classid]
        assert classid in self.baseRegistry

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def registerClass(self, derived):
        """
        Register a derived class, i.e. a class deriving from a registered
        base class in the configuration system.
        :param derived:
        :return:
        """
        self.baseRegistry[derived.__classid__] = derived
        return self

    def getSchema(self, classid, rules=AssemblyRules()):
        """
        Get schema for class with "classid" derived from base class given to
        constructor using assembly "rules".
        Example:
                schema = Configurator(Shape).getSchema("Rectangle")
        """
        if rules is None:
            rules = AssemblyRules()
        if isinstance(classid, type):
            classid = classid.__classid__
        if not isinstance(classid, str):
            raise TypeError("The first argument type '{}' is not allowed. "
                            "Must be a type or str.".format(type(classid)))
        if classid not in self.baseRegistry:
            raise AttributeError("Class Id '{}' not found in the base registry"
                                 .format(classid))
        Derived = self.baseRegistry[classid]  # userclass -> Derived

        # building list of classes in inheritance order from bases to the
        # last derived
        def inheritanceChain(c, bases_id, clist):
            if not isinstance(c, type):
                return
            classId = getattr(c, '__classid__', None)
            if classId is not None and classId not in bases_id:
                for x in c.__bases__:
                    inheritanceChain(x, bases_id, clist)
            if c not in clist:
                clist.append(c)

        clist = []
        inheritanceChain(Derived, Derived.__bases_classid__, clist)
        # clist contains list of classes in inheritance order
        schema = Schema(classid, rules)
        for theClass in clist:
            if hasattr(theClass, "expectedParameters"):
                # fill schema in order from base to derived
                theClass.expectedParameters(schema)

        return schema

    def create(self, *args):
        """
        The factory method to create the instance of class with "classId" that
        inherits from base class given to constructor using "input"
        configuration.
        The last argument is a flag to determine if the input configuration
        should be validated.

        Example:

                instance = Configurator(Shape).create("EditableCircle",
                                                      Hash("radius", 12.345))

        The factory method to create instance of class that inherits from base
        class given to constructor using input "configuration".
        The configuration should have "classId" of class to be created as a
        root element.  The last argument is a flag to determine if the input
        "configuration" should be validated.

        Example:

                configuration = Hash("EditableCircle.radius", 12.345)
                instance = Configurator(Shape).create(configuration)

        :param args: thus can have an arity of one to three:
                    - class id string
                    - config with classId as root key
                    - class id string, config
                    - config with classId as root key, validation flag
                    - class id string, config, validation flag

        """

        if len(args) == 3:
            classid = args[0]
            configuration = args[1]
            validation = args[2]
        elif len(args) == 2:
            if isinstance(args[1], bool):
                configuration = args[0]
                validation = args[1]
                classid = list(configuration.keys())[0]
                configuration = configuration[classid]
            else:
                classid = args[0]
                configuration = args[1]
                validation = True
        elif len(args) == 1:
            configuration = args[0]
            validation = True
            classid = list(configuration.keys())[0]
            configuration = configuration[classid]
        else:
            raise TypeError("Wrong number of arguments and/or their types")
        if isinstance(classid, type):
            classid = classid.__classid__
        if not isinstance(classid, str):
            raise TypeError("First argument 'classid' must be a python"
                            " string type")
        if classid not in self.baseRegistry:
            raise AttributeError("Unknown classid '{}' in base registry"
                                 .format(classid))
        Derived = self.baseRegistry[classid]

        schema = Configurator(Derived.__base_classid__).getSchema(
            classid, AssemblyRules())
        if not validation:
            return Derived(configuration)
        validator = Validator()
        result, error, validated = validator.validate(schema, configuration)
        if not result:
            raise RuntimeError(f"Validation Exception: {error}")
        return Derived(validated)

    def createNode(self, nodename, classid, configuration, validation=True):
        """
        The helper method to create instance of class specified by "classId"
        and derived from class given to constructor using sub-configuration
        specified by "nodeName" which has to be a part of input
        "configuration".

        The last argument is a flag to determine if the input "configuration"
        should be validated.
        """

        if isinstance(classid, type):
            classid = classid.__classid__
        if not isinstance(classid, str):
            raise TypeError("Second argument 'classid' must be a python"
                            " string type")
        if nodename in configuration:
            return self.create(classid, configuration[nodename], validation)
        raise AttributeError('Given nodeName "{}" is not part of input'
                             ' configuration'.format(nodename))

    def getRegisteredClasses(self):
        """
        Returns list of "classid"'s for all registered classes derived from
        base class given to constructor.
        """
        return list(self.baseRegistry.keys())

    @staticmethod
    def getRegisteredBaseClasses():
        """
        Returns all classid's of base classes registered in Configurator.
        """
        return list(Configurator.registry.keys())



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

import weakref

import numpy as np

from karabo.bound import BinarySerializerHash, Hash, ImageData, fullyEqual


def test_ndarray_io():

    ser = BinarySerializerHash.create('Bin')

    def func(i):
        arr = np.full((100, 200), 88+1, dtype=np.int16)
        data = Hash('node', Hash())
        node = data['node']
        node.set('ndarray', arr)
        node.set('index', i)
        archive = ser.save(data)
        dcopy = ser.load(archive)
        assert fullyEqual(dcopy, data)

    for i in range(100):
        func(i)


def test_ndarray_refcount():
    arr = np.arange(20000, dtype=np.int16).reshape(100, 200)
    arr_weak = weakref.ref(arr)
    h = Hash('a', arr)
    g = h['a']
    img = ImageData(arr)
    assert arr_weak() is not None
    del arr
    assert arr_weak() is not None
    del h
    assert arr_weak() is not None
    del g
    assert arr_weak() is not None
    del img
    assert arr_weak() is None



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa: F401
from karabind import (
    ALARM_ELEMENT, ATTOSEC, BOOL_ELEMENT, BYTEARRAY_ELEMENT, DOUBLE_ELEMENT,
    FEMTOSEC, FLOAT_ELEMENT, IMAGEDATA_ELEMENT, INPUT_CHANNEL, INT32_ELEMENT,
    INT64_ELEMENT, MICROSEC, MILLISEC, NANOSEC, NDARRAY_ELEMENT, NODE_ELEMENT,
    OUTPUT_CHANNEL, OVERWRITE_ELEMENT, PICOSEC, SLOT_ELEMENT, STATE_ELEMENT,
    STRING_ELEMENT, TABLE_ELEMENT, TIME_UNITS, UINT32_ELEMENT, UINT64_ELEMENT,
    VECTOR_BOOL_ELEMENT, VECTOR_CHAR_ELEMENT, VECTOR_DOUBLE_ELEMENT,
    VECTOR_FLOAT_ELEMENT, VECTOR_INT32_ELEMENT, VECTOR_INT64_ELEMENT,
    VECTOR_STRING, VECTOR_STRING_ELEMENT, VECTOR_UINT32_ELEMENT,
    VECTOR_UINT64_ELEMENT, AccessLevel, AccessType, ArchivePolicy,
    AssemblyRules, AssignmentType as Assignment, BinarySerializerHash,
    BinarySerializerSchema, Broker, Category, Channel, ChannelMetaData,
    ClassInfo, Connection, ConnectionStatus, DaqDataType, DateTimeString,
    DeviceClient as BoundDeviceClient, Dims, Encoding, Epochstamp, ErrorCode,
    EventLoop, Hash, HashAttributes, HashAttributesNode, HashFilter,
    HashMergePolicy, HashNode, ImageData, InputChannel, InputHash, InputSchema,
    Logger, MetricPrefix, NodeType, OutputChannel, OutputHash, OutputSchema,
    PriorityLevel, Requestor, Rotation, Schema, SignalSlotable,
    SignalSlotableIntern, Slot, SlotElementBase, TextSerializerHash,
    TextSerializerSchema, TimeDuration, TimeId, Timestamp, Types, TypesClass,
    Unit, Validator, ValidatorValidationRules, VectorHash, VectorHashPointer,
    _DimsIntern, cppNDArray, cppNDArrayCopy, fullyEqual, generateAutoStartHash,
    jsonToHash, loadFromFile, loadHashFromFile, loadSchemaFromFile,
    saveHashToFile, saveSchemaToFile, saveToFile, setDims, similar,
    startDeviceServer, stopDeviceServer)

from ..common.alarm_conditions import AlarmCondition
from ..common.states import State, StateSignifier
from .configurator import Configurator
from .decorators import KARABO_CLASSINFO, KARABO_CONFIGURATION_BASE_CLASS
from .device import PythonDevice
from .device_client import DeviceClient
from .device_server import DeviceServer, Launcher
from .plugin_loader import PluginLoader
from .runner import Runner
from .worker import Worker

# For comptibility with old karathon bindings, take care that e.g.
# str(Types.INT32) is 'INT32' and not 'Types.INT32'


def __typesToString(self):
    repStr = repr(self)  # i.e. '<Types.{this we want}: {enum value}>'
    return repStr[7:repStr.find(':')]


setattr(Types, "__str__", __typesToString)


class AbstractInput:
    pass


def isStdVectorDefaultConversion():
    return True


def setStdVectorDefaultConversion():
    pass



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import unittest

from karabo.bound import (
    Hash, InputSchema, OutputSchema, Schema, loadFromFile, saveToFile, similar)

from .configuration_example_classes import TestStruct1


class Io_TestCase(unittest.TestCase):
    def setUp(self):
        self.schema = Schema()
        TestStruct1.expectedParameters(self.schema)

    def tearDown(self):
        del self.schema
        self.schema = None

    def test_io1_write_xml_schema(self):
        try:
            out = OutputSchema.create("TextFile",
                                      Hash("filename", "/tmp/testschema.xml",
                                           "format", "Xml",
                                           "Xml.indentation", 3))
            self.assertIsNotNone(out)
            out.write(self.schema)
            del out

            saveToFile(self.schema, "/tmp/testschema2.xml",
                       Hash("format", "Xml", "Xml.indentation", 3))
        except Exception as e:
            self.fail("test_io_write_xml_schema exception group: " + str(e))

    def test_io2_read_xml_schema(self):
        try:
            inp = InputSchema.create("TextFile",
                                     Hash("filename", "/tmp/testschema.xml"))
            self.assertIsNotNone(inp)
            schema1 = Schema()
            inp.read(schema1)
            del inp

            schema2 = Schema()
            loadFromFile(schema2, "/tmp/testschema2.xml")

            self.assertTrue(similar(schema1, self.schema))
            self.assertTrue(similar(schema2, self.schema))
        except Exception as e:
            self.fail("test_io_read_xml_schema exception group: " + str(e))

    def test_io1_write_binary_schema(self):
        try:
            out = OutputSchema.create("BinaryFile",
                                      Hash("filename", "/tmp/testschema.bin"))
            self.assertIsNotNone(out)
            out.write(self.schema)
            del out

            saveToFile(self.schema, "/tmp/testschema2.bin")
        except Exception as e:
            self.fail("test_io_write_binary_schema exception group: " + str(e))

    def test_io2_read_binary_schema(self):
        try:
            inp = InputSchema.create("BinaryFile",
                                     Hash("filename", "/tmp/testschema.bin"))
            self.assertIsNotNone(inp)
            schema1 = Schema()
            inp.read(schema1)
            del inp

            schema2 = Schema()
            loadFromFile(schema2, "/tmp/testschema2.bin")

            self.assertTrue(similar(schema1, self.schema))
            self.assertTrue(similar(schema2, self.schema))
        except Exception as e:
            self.fail("test_io_read_binary_schema exception group: " + str(e))


if __name__ == '__main__':
    unittest.main()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import traceback
from importlib.metadata import entry_points

from karabind import STRING_ELEMENT

from .decorators import KARABO_CLASSINFO, KARABO_CONFIGURATION_BASE_CLASS

DEFAULT_NAMESPACE = "karabo.bound_device"


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("PythonPluginLoader", "1.0")
class PluginLoader:

    @staticmethod
    def expectedParameters(expected):

        e = STRING_ELEMENT(expected).key("pluginNamespace")
        e.displayedName("Plugin Namespace")
        e.description("Namespace to search for plugins")
        e.assignmentOptional().defaultValue(DEFAULT_NAMESPACE)
        e.expertAccess().commit()

    def __init__(self, config):
        self._entrypoints = []
        self.plugin_errors = {}

        if "pluginNamespace" in config:
            self.pluginNamespace = config["pluginNamespace"]
        else:
            msg = "A namespace must be defined for plugins to load."
            raise ValueError(msg)

    def getPlugin(self, name):
        for ep in self._entrypoints:
            if ep.name == name:
                return ep
        else:
            raise RuntimeError(f"Plugin {name} not found!")

    def update(self):
        """Protected entry to update the entrypoints"""
        def load(ep):
            try:
                ep.load()
                return True
            except BaseException:
                self.plugin_errors[ep.name] = traceback.format_exc()
                return False

        self._entrypoints = [
            ep for ep in entry_points(group=self.pluginNamespace)
            if load(ep)]

        return self._entrypoints



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import json
from zlib import adler32

from karabo.bound import BinarySerializerSchema
from karabo.bound.testing import ServerContext, eventLoop, sleepUntil


def test_adler():
    s = b'I love python, Hello world'
    t = adler32(s)
    assert t == 2073102698, "The adler32's behavior changed."


def test_schema(eventLoop: eventLoop):
    config = {"AdlerTestDevice": {"classId": "TestDevice",
                                  "middlelayerDevice": "Notthere"}}
    server = ServerContext(
        "testAdler",
        [f"init={json.dumps(config)}",
         "pluginNamespace=karabo.bound_device_test"])
    with server:
        remote = server.remote()
        sleepUntil(lambda: "AdlerTestDevice" in remote.getDevices(),
                   timeout=10)
        schema = remote.getDeviceSchema("AdlerTestDevice")
        ser = BinarySerializerSchema.create("Bin")
        schema = ser.save(schema)
        text = ("The generated schema changed. If this "
                "is desired, change the checksum in the code.")
        checksum = adler32(schema)
        assert checksum == 3617127536, text



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

import time

import pytest

from karabo.bound import Worker


def test_worker_config():
    worker = Worker(timeout=2000)
    worker.start()
    time.sleep(0.1)
    assert not worker.running

    # Threads can be only started once
    worker.set(callback=lambda x: x)
    with pytest.raises(RuntimeError):
        worker.start()

    # test order of arguments
    worker = Worker(None, 2000, 1)
    assert worker.timeout == 2000
    assert worker.single_shot


def test_worker_basic():
    worker = Worker(timeout=2000)
    worker.setTimeout(100)
    assert not worker.is_running()
    assert worker.timeout == 100

    counter = 0

    def increase():
        nonlocal counter
        counter += 1

    assert not worker.is_running()
    worker.set(callback=increase, timeout=50)
    assert worker.timeout == 50
    worker.start()
    assert worker.is_running()
    while counter < 5:
        time.sleep(0.01)
    assert counter == 5
    worker.pause()
    # We can pause
    time.sleep(0.5)
    assert counter == 5
    worker.start()
    while counter < 10:
        time.sleep(0.01)
    worker.stop()
    assert counter == 10
    assert not worker.is_running()

    # Can't set while running
    worker = Worker()
    worker.set(callback=increase, timeout=50)
    worker.start()
    with pytest.raises(RuntimeError):
        worker.set(callback=increase, timeout=50)
    worker.stop()


def test_worker_singleshot():
    counter = 0

    def increase():
        nonlocal counter
        counter += 1

    worker = Worker(increase, 100, single_shot=True)
    assert worker.timeout == 100
    worker.start()
    time.sleep(0.03)
    assert counter == 0
    time.sleep(0.1)
    assert counter == 1
    assert not worker.running
    time.sleep(0.15)
    assert counter == 1



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

from karabind import DeviceClient as BoundDeviceClient, Hash
from karabo.common.alarm_conditions import AlarmCondition
from karabo.common.api import KARABO_CLASS_ID_ALARM, KARABO_CLASS_ID_STATE
from karabo.common.states import State


class DeviceClient(BoundDeviceClient):
    """ Handle certain device client responsibilities which interact with other
    Python code in Karabo.
    """

    def get(self, instanceId, *args, **kw):
        """ Handle conversion of returned State and AlarmCondition objects."""
        value = super().get(instanceId, *args, **kw)

        if not isinstance(value, Hash):
            schema = self.getDeviceSchema(instanceId)
            key = args[0]
            if schema.isLeaf(key):
                if schema.hasClassId(key):
                    classId = schema.getClassId(key)
                    if classId == KARABO_CLASS_ID_STATE:
                        return State(value)
                    elif classId == KARABO_CLASS_ID_ALARM:
                        return AlarmCondition(value)

        return value
